from pyqtgraph.parametertree.parameterTypes import GroupParameterItem
from pyqtgraph import mkPen
from PyQt5 import QtCore
from PyQt5.QtWidgets import QMenu

translate = QtCore.QCoreApplication.translate

from pymead.gui.custom_context_menu_event import custom_context_menu_event


class SelectableHeaderParameterItem(GroupParameterItem):

    def __init__(self, param, depth):
        super().__init__(param, depth)

    def selected(self, sel):
        if self.param.parent() is None:
            return
        if sel:
            self.param.parent().status_bar.showMessage(f"Airfoil {self.param.name()} selected", 3000)
            for curve in self.param.parent().mea.airfoils[self.param.name()].curve_list:
                curve.pg_curve_handle.setPen(mkPen(color='fuchsia', width=2))
        else:
            for curve in self.param.parent().mea.airfoils[self.param.name()].curve_list:
                curve.pg_curve_handle.setPen(mkPen(color='cornflowerblue', width=2))

    def contextMenuEvent(self, ev):
        custom_context_menu_event(ev, self)





def calculate_normal_shock_total_pressure_ratio(M_up: float, gam: float) -> float:
    """
    Normal-shock relation that calculates the total pressure ratio across a shock (:math:`p_{t,y} / p_{t,x}` ,
    where :math:`x` is the state immediately upstream of the shock, and :math:`y` is the state immediately downstream
    of the shock).

    Parameters
    ==========
    M_up: float
        Mach number immediately upstream of the shock and normal to the shock (:math:`M_x`)

    gam: float
        Specific heat ratio

    Returns
    =======
    float
        Total pressure ratio across the shock wave, :math:`p_{t,y} / p_{t,x}`
    """
    A = (gam + 1) / 2 * M_up**2
    B = 1 + (gam - 1) / 2 * M_up**2
    C = 2 * gam * M_up**2 / (gam + 1)
    D = (gam - 1) / (gam + 1)
    return (A / B) ** (gam / (gam - 1)) * (C - D) ** (1 / (1 - gam))



from copy import deepcopy

import pyqtgraph as pg
from PyQt5.QtWidgets import QWidget, QDoubleSpinBox, QSpinBox, QLabel
from PyQt5.QtWidgets import QGridLayout

from pymead.core.geometry_collection import GeometryCollection
from pymead.optimization.sampling import ConstrictedRandomSampling


class SamplingVisualizationWidget(QWidget):
    def __init__(self, parent, geo_col_dict: dict, initial_sampling_width: float, initial_n_samples: int,
                 background_color: str):
        super().__init__(parent=parent)

        self.geo_col_dict = deepcopy(geo_col_dict)
        self.geo_col_dict["airfoil_graphs_active"] = False
        self.geo_col = GeometryCollection.set_from_dict_rep(self.geo_col_dict)
        norm_param_list = self.geo_col.extract_design_variable_values(bounds_normalized=True)
        self.norm_param_list = deepcopy(norm_param_list)

        self.w = pg.GraphicsLayoutWidget(show=True, size=(800, 300))
        # self.w.setWindowTitle('Airfoil')
        self.w.setBackground(background_color)

        self.v = self.w.addPlot()
        self.v.setLabel(axis="bottom", text="x")
        self.v.setLabel(axis="left", text="y")
        self.v.setAspectLocked(True)

        layout = QGridLayout(self)
        layout.addWidget(self.w, 0, 0, 1, 4)

        self.spinbox_ws = QDoubleSpinBox(self)
        self.spinbox_ns = QSpinBox(self)

        self.spinbox_ws.setMinimum(0.0)
        self.spinbox_ws.setMaximum(1.0)
        self.spinbox_ws.setDecimals(4)
        self.spinbox_ws.setSingleStep(0.01)
        self.spinbox_ws.setValue(initial_sampling_width)

        self.spinbox_ns.setMaximum(1000)
        self.spinbox_ns.setValue(initial_n_samples)

        self.spinbox_ws.valueChanged.connect(self.spinbox_ws_value_changed)
        self.spinbox_ns.valueChanged.connect(self.spinbox_ns_value_changed)

        self.slider_ws_label = QLabel(f"Sampling Width:", self)
        self.slider_ns_label = QLabel(f"Number of Samples:", self)

        layout.addWidget(self.slider_ws_label, 1, 0, 1, 1)
        layout.addWidget(self.spinbox_ws, 1, 1, 1, 1)
        layout.addWidget(self.slider_ns_label, 1, 2, 1, 1)
        layout.addWidget(self.spinbox_ns, 1, 3, 1, 1)

        self.update_plots(sampling_width=initial_sampling_width, n_samples=initial_n_samples)

        self.current_ws = initial_sampling_width
        self.current_ns = initial_n_samples

        self.setLayout(layout)
        self.setMinimumSize(800, 300)

    def spinbox_ws_value_changed(self, new_ws):
        self.current_ws = new_ws
        self.update_plots(sampling_width=self.current_ws, n_samples=self.current_ns)

    def spinbox_ns_value_changed(self, new_ns):
        self.current_ns = new_ns
        self.update_plots(sampling_width=self.current_ws, n_samples=self.current_ns)

    def update_plots(self, sampling_width: float, n_samples: int):
        self.v.clear()
        sampling = ConstrictedRandomSampling(n_samples=n_samples, norm_param_list=deepcopy(self.norm_param_list),
                                             max_sampling_width=sampling_width)
        X_list = sampling.sample()
        for idx, individual in enumerate(X_list):
            self.geo_col.assign_design_variable_values(individual, bounds_normalized=True)
            for a in self.geo_col.container()["airfoils"].values():
                coords = a.get_coords_selig_format()
                plot_handle = self.v.plot(pen=pg.mkPen(color="cornflowerblue"))
                plot_handle.setData(coords[:, 0], coords[:, 1])



import numpy as np

from pymead.core.parametric_curve import ParametricCurve
from pymead.core.bezier import Bezier


class RationalBSpline(ParametricCurve):
    """This code was produced in part by the ChatGPT bot on openai.com"""

    def get_curvature_comb(self, max_k_normalized_scale_factor, interval: int = 1):
        pass

    def __init__(self, P, knots, weights):
        self.P = np.array(P)
        self.knots = knots
        self.weights = weights
        t = np.linspace(0.0, 0.999, 101)
        curve = self.evaluate_vector(t)
        super().__init__(t=t, x=curve[:, 0], y=curve[:, 1], px=None, py=None, ppx=None, ppy=None, R=None, k=None)

    def evaluate(self, t):
        n = len(self.P) - 1
        m = len(self.knots) - 1
        p = m - n - 1

        # Find the span for the given value of t
        span = self._find_span(n, p, t)

        # Compute the basis functions
        basis_functions = self._basis_functions(span, t, p)

        # Evaluate the curve at t
        curve_point = np.zeros(self.P.shape[1])
        for i in range(p + 1):
        # for i in range(n + 1):
            curve_point += basis_functions[i] * self.P[span - p + i] * self.weights[span - p + i]
            # curve_point += self._basis_function(i, p, t) * self.P[i]

        # curve_point2 = self.de_boor(t)
        return curve_point

    def evaluate_vector(self, t_vec):
        """Evaluate the curve along a specified parameter vector"""
        points = []
        for t_val in t_vec:
            point = self.evaluate(t_val)
            points.append(point)
        return np.array([self.evaluate(t_val) for t_val in t_vec])
        # return np.array([self.de_boor(t_val) for t_val in t_vec])

    def _find_span(self, n, p, t):
        """Find the span for the given value of t"""
        # if t == self.knots[n + 1]:
        #     return n
        # low, high = p, n + 1
        # mid = (low + high) // 2
        # while t < self.knots[mid] or t >= self.knots[mid + 1]:
        #     if t < self.knots[mid]:
        #         high = mid
        #     else:
        #         low = mid
        #     mid = (low + high) // 2
        #     print(f"{mid = }")
        # return mid
        interval = 0
        for i in range(len(self.knots) - 1):
            # Modification to account for repeated knots
            if np.isclose(self.knots[i + 1] - self.knots[i], 0.0):
                continue
            if self.knots[i] <= t <= self.knots[i + 1]:
                interval = i
                break
        return interval

    # def _basis_function(self, i: int, p: int, t: float):
    #     """Implementation of the Cox-de Boor recursion formula
    #     (see `this page <https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/B-spline/bspline-basis.html>`_)"""
    #
    #     def cox_de_boor_recursion(i_, p_):
    #         if p_ == 0:
    #             return 1 if self.knots[i_] <= t < self.knots[i_ + 1] else 0
    #         else:
    #             return (t - self.knots[i_]) / (self.knots[i_ + p_] - self.knots[i_]) * \
    #                    cox_de_boor_recursion(i_, p_ - 1) + \
    #                    (self.knots[i_ + p_ + 1] - t) / (self.knots[i_ + p_ + 1] - self.knots[i_ + 1]) * \
    #                    cox_de_boor_recursion(i_ + 1, p_ - 1)
    #
    #     return cox_de_boor_recursion(i, p)

    def de_boor(self, t):
        """From wikpedia and https://stackoverflow.com/questions/57507696/b-spline-derivative-using-de-boors-algorithm"""
        # Find the knot interval that contains t
        interval = 0
        for i in range(len(self.knots) - 1):
            # Modification to account for repeated knots
            if np.isclose(self.knots[i + 1] - self.knots[i], 0.0):
                continue
            if self.knots[i] <= t <= self.knots[i + 1]:
                interval = i
                break

        p = len(self.knots) - self.P.shape[0] - 1

        d = [self.P[j + interval - p] for j in range(0, p + 1)]

        for r in range(1, p + 1):
            for j in range(p, r - 1, -1):
                print(f"{r = }, {j = }, {p = }, {interval = }")
                alpha = (t - self.knots[j + interval - p]) / (self.knots[j + 1 + interval - r] - self.knots[j + interval - p])
                d[j] = (1.0 - alpha) * d[j - 1] + alpha * d[j]

        return d[p]

    def _basis_functions(self, span, t, p):
        """Compute the non-vanishing basis functions"""
        basis_functions = np.zeros(p + 1)
        basis_functions[0] = 1.0
        left = np.zeros(p + 1)
        right = np.zeros(p + 1)
        for j in range(1, p + 1):
            left[j] = t - self.knots[span + 1 - j]
            right[j] = self.knots[span + j] - t
            saved = 0.0
            for r in range(j):
                temp = basis_functions[r] / (right[r + 1] + left[j - r])
                basis_functions[r] = saved + right[r + 1] * temp
                saved = left[j - r] * temp
            basis_functions[j] = saved
        return basis_functions


class BSpline(RationalBSpline):
    def __init__(self, P, knots):
        weights = np.ones(P.shape[0])
        super().__init__(P=P, knots=knots, weights=weights)


class ClampedBSpline(BSpline):
    def __init__(self, P: np.ndarray, extra_knots: np.ndarray):
        n = P.shape[0] - 1
        multiplicity = n - len(extra_knots) + 1
        knots = np.concatenate((np.zeros(multiplicity), extra_knots, np.ones(multiplicity)))
        super().__init__(P=P, knots=knots)


class BezierCurve(ClampedBSpline):
    def __init__(self, P):
        super().__init__(P=P, extra_knots=np.array([]))


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    P = np.array([[0.0, 0.0], [0.3, 0.1], [0.5, 0.0], [0.7, -0.1], [1.0, 0.0]])
    b = Bezier(P=P, t=np.linspace(0, 1, 101))
    # b2 = BezierCurve(P=P)
    # knots = np.array([0, 0, 0, 0.3, 0.7, 1, 1, 1])
    # knots = np.linspace(0.0, 1.0, 10)
    knots = np.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
    plt.plot(b.x, b.y)
    # plt.plot(b2.x, b2.y)
    # plt.show()
    bspline = BSpline(P=P, knots=knots)
    plt.plot(bspline.x, bspline.y)
    P = np.array([[0.0, 0.0], [0.3, 0.1], [0.5, 0.0], [0.7, -0.1], [1.0, 0.0]])
    # knots = np.linspace(0, 1, 14)
    # bspline = BSpline(P=P, knots=knots)
    # plt.plot(bspline.x, bspline.y)
    plt.show()
    pass



from collections import namedtuple

import numpy as np


def measure_distance(x1: float, y1: float, x2: float, y2: float):
    return np.hypot(x1 - x2, y1 - y2)


def measure_abs_angle(x1: float, y1: float, x2: float, y2: float):
    return (np.arctan2(y2 - y1, x2 - x1)) % (2 * np.pi)


def measure_rel_angle3(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return (np.arctan2(y1 - y2, x1 - x2) - np.arctan2(y3 - y2, x3 - x2)) % (2 * np.pi)


def measure_rel_angle4(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float):
    return (np.arctan2(y4 - y3, x4 - x3) - np.arctan2(y2 - y1, x2 - x1)) % (2 * np.pi)


def measure_point_line_distance_signed(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    """
    Measures the signed distance from a point :math:`(x_3, y_3)` to a line defined by
    two points: :math:`(x_1, y_1)` and :math:`(x_2, y_2)`. This signed version is used in the point-line
    symmetry constraint because a perpendicular constraint does not necessarily prevent the GCS from moving
    the symmetric point to the same side of the line as the target point.

    Parameters
    ----------
    x1: float
        x-coordinate of the line's start point
    y1: float
        y-coordinate of the line's start point
    x2: float
        x-coordinate of the line's end point
    y2: float
        y-coordinate of the line's end point
    x3: float
        x-coordinate of the target point
    y3: float
        y-coordinate of the target point

    Returns
    -------
    float
        Distance from the target point to the line
    """
    return ((x2 - x1) * (y1 - y3) - (x1 - x3) * (y2 - y1)) / measure_distance(x1, y1, x2, y2)


def measure_point_line_distance_unsigned(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return np.abs(measure_point_line_distance_signed(x1, y1, x2, y2, x3, y3))


def measure_radius_of_curvature_bezier(Lt: float, Lc: float, n: int, psi: float):
    return np.abs(np.true_divide(Lt ** 2, Lc * (1 - 1 / n) * np.sin(psi)))


def measure_curvature_length_bezier(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, R: float, n: int):
    Lt = measure_distance(x1, y1, x2, y2)
    psi = measure_rel_angle3(x1, y1, x2, y2, x3, y3)
    with np.errstate(divide="ignore"):
        Lc = np.abs(np.true_divide(Lt ** 2, R * (1 - 1 / n) * np.sin(psi)))
    return Lc


def measure_curvature_bezier(Lt: float, Lc: float, n: int, psi: float):
    with np.errstate(divide="ignore"):
        kappa = np.abs(np.true_divide(Lc * (1 - 1 / n) * np.sin(psi), Lt ** 2))
    return kappa


def measure_data_bezier_curve_joint(xy: np.ndarray, n: np.ndarray):
    phi1 = measure_abs_angle(xy[2, 0], xy[2, 1], xy[1, 0], xy[1, 1])
    phi2 = measure_abs_angle(xy[2, 0], xy[2, 1], xy[3, 0], xy[3, 1])
    theta1 = measure_abs_angle(xy[1, 0], xy[1, 1], xy[0, 0], xy[0, 1])
    theta2 = measure_abs_angle(xy[3, 0], xy[3, 1], xy[4, 0], xy[4, 1])
    psi1 = theta1 - phi1
    psi2 = theta2 - phi2
    phi_rel = (phi1 - phi2) % (2 * np.pi)
    Lt1 = measure_distance(xy[1, 0], xy[1, 1], xy[2, 0], xy[2, 1])
    Lt2 = measure_distance(xy[2, 0], xy[2, 1], xy[3, 0], xy[3, 1])
    Lc1 = measure_distance(xy[0, 0], xy[0, 1], xy[1, 0], xy[1, 1])
    Lc2 = measure_distance(xy[3, 0], xy[3, 1], xy[4, 0], xy[4, 1])
    kappa1 = measure_curvature_bezier(Lt1, Lc1, n[0], psi1)
    kappa2 = measure_curvature_bezier(Lt2, Lc2, n[1], psi2)
    with np.errstate(divide="ignore"):
        R1 = np.true_divide(1, kappa1)
        R2 = np.true_divide(1, kappa2)
    n1 = n[0]
    n2 = n[1]
    field_names = ["phi1", "phi2", "theta1", "theta2", "psi1", "psi2", "phi_rel", "Lt1", "Lt2", "Lc1", "Lc2",
                   "kappa1", "kappa2", "R1", "R2", "n1", "n2"]
    BezierCurveJointData = namedtuple("BezierCurveJointData", field_names=field_names)
    data = BezierCurveJointData(phi1=phi1, phi2=phi2, theta1=theta1, theta2=theta2, psi1=psi1, psi2=psi2,
                                phi_rel=phi_rel, Lt1=Lt1, Lt2=Lt2, Lc1=Lc1, Lc2=Lc2, kappa1=kappa1, kappa2=kappa2,
                                R1=R1, R2=R2, n1=n1, n2=n2)
    return data


def radius_of_curvature_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, R: float, n: int):
    Lt = measure_distance(x1, y1, x2, y2)
    psi = measure_rel_angle3(x1, y1, x2, y2, x3, y3)
    Lc = np.abs(np.true_divide(Lt ** 2, R * (1 - 1 / n) * np.sin(psi)))
    return distance_constraint(x2, y2, x3, y3, Lc)


def empty_constraint_weak():
    return 0.0


def fixed_param_constraint(p_val: float, val: float):
    return p_val - val


def fixed_param_constraint_weak(new_val: float, old_val: float):
    return new_val - old_val


def fixed_x_constraint(x: float, val: float):
    return x - val


def fixed_x_constraint_weak(x_new: float, x_old: float):
    return x_new - x_old


def fixed_y_constraint(y: float, val: float):
    return y - val


def fixed_y_constraint_weak(y_new: float, y_old: float):
    return y_new - y_old


def distance_constraint(x1: float, y1: float, x2: float, y2: float, dist: float):
    return measure_distance(x1, y1, x2, y2) - dist


def distance_constraint_weak(x1_new: float, y1_new: float, x2_new: float, y2_new: float,
                             x1_old: float, y1_old: float, x2_old: float, y2_old: float):
    return (measure_distance(x1_new, y1_new, x2_new, y2_new) -
            measure_distance(x1_old, y1_old, x2_old, y2_old))


def abs_angle_constraint(x1: float, y1: float, x2: float, y2: float, angle: float):
    return measure_abs_angle(x1, y1, x2, y2) - angle


def abs_angle_constraint_weak(x1_new: float, y1_new: float, x2_new: float, y2_new: float,
                              x1_old: float, y1_old: float, x2_old: float, y2_old: float):
    return (measure_abs_angle(x1_new, y1_new, x2_new, y2_new) -
            measure_abs_angle(x1_old, y1_old, x2_old, y2_old))


def rel_angle3_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, angle: float):
    return measure_rel_angle3(x1, y1, x2, y2, x3, y3) - angle


def rel_angle3_constraint_weak(x1_new: float, y1_new: float, x2_new: float, y2_new: float, x3_new: float, y3_new: float,
                               x1_old: float, y1_old: float, x2_old: float, y2_old: float, x3_old: float,
                               y3_old: float):
    return (measure_rel_angle3(x1_new, y1_new, x2_new, y2_new, x3_new, y3_new) -
            measure_rel_angle3(x1_old, y1_old, x2_old, y2_old, x3_old, y3_old))


def rel_angle4_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float,
                          angle: float):
    return measure_rel_angle4(x1, y1, x2, y2, x3, y3, x4, y4) - angle


def perp3_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return measure_rel_angle3(x1, y1, x2, y2, x3, y3) - (np.pi / 2)


def perp4_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float):
    return measure_rel_angle4(x1, y1, x2, y2, x3, y3, x4, y4) - (np.pi / 2)


def antiparallel3_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return measure_rel_angle3(x1, y1, x2, y2, x3, y3) - np.pi


def antiparallel4_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float):
    return measure_rel_angle4(x1, y1, x2, y2, x3, y3, x4, y4) - np.pi


def parallel3_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return measure_rel_angle3(x1, y1, x2, y2, x3, y3)


def parallel4_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float):
    return measure_rel_angle4(x1, y1, x2, y2, x3, y3, x4, y4)


def point_on_line_constraint(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float):
    return (y1 - y2) * (x3 - x1) - (y3 - y1) * (x1 - x2)


def points_equidistant_from_line_constraint_unsigned(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float,
                                                     x4: float, y4: float):
    return (measure_point_line_distance_unsigned(x1, y1, x2, y2, x3, y3) -
            measure_point_line_distance_unsigned(x1, y1, x2, y2, x4, y4))


def points_equidistant_from_line_constraint_signed(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float,
                                                   x4: float, y4: float):
    return (measure_point_line_distance_signed(x1, y1, x2, y2, x3, y3) +
            measure_point_line_distance_signed(x1, y1, x2, y2, x4, y4))



import itertools
import os
import typing

import numpy as np
from pymead.utils.read_write_files import save_data

from pymead.core import UNITS
from pymead.core.transformation import Transformation2D
from pymead.core.airfoil import Airfoil
from pymead.core.pymead_obj import PymeadObj
from pymead.plugins.IGES.curves import BezierIGES
from pymead.plugins.IGES.iges_generator import IGESGenerator


class MEA(PymeadObj):

    def __init__(self, airfoils: typing.List[Airfoil], name: str or None = None):
        self.airfoils = airfoils
        self.reference_airfoil = self.airfoils[0]
        super().__init__(sub_container="mea")

        # Name the MEA
        name = "MEA-1" if name is None else name
        self._name = None
        self.set_name(name)

    def add_airfoil(self, airfoil: Airfoil):
        self.airfoils.append(airfoil)

    def remove_airfoil(self, airfoil: Airfoil):
        self.airfoils.remove(airfoil)

    def get_coords_list(self, max_airfoil_points: int = None, curvature_exp: float = 2.0):
        mea_coords_list = [
            airfoil.get_coords_selig_format(max_airfoil_points, curvature_exp) for airfoil in self.airfoils]
        max_y = [np.max(coords[:, 1]) for coords in mea_coords_list]
        airfoil_order = np.argsort(max_y)[::-1]
        return [mea_coords_list[a_idx] for a_idx in airfoil_order]

    def get_coords_list_chord_relative(self, max_airfoil_points: int = None,
                                       curvature_exp: float = 2.0):
        coords_list = self.get_coords_list(max_airfoil_points=max_airfoil_points, curvature_exp=curvature_exp)

        # Get the transformation object
        chord_length = self.reference_airfoil.measure_chord()
        transformation_kwargs = dict(
            tx=[0.0], ty=[0.0], r=[0.0], sx=[1 / chord_length], sy=[1 / chord_length],
            rotation_units="rad", order="t,s,r"
        )
        transformation = Transformation2D(**transformation_kwargs)
        transformation_kwargs["length_unit"] = UNITS.current_length_unit()

        return [transformation.transform(coords) for coords in coords_list], transformation_kwargs

    def write_mses_blade_file(self,
                              airfoil_sys_name: str,
                              blade_file_dir: str,
                              mea_coords_list: typing.List[np.ndarray] or None = None,
                              grid_bounds: typing.List[float] or None = None,
                              max_airfoil_points: int = None, curvature_exp: float = 2.0) -> (str, typing.List[str]):

        # Get the MEA coordinates list if not provided
        if mea_coords_list is None:
            mea_coords_list, transformation_kwargs = self.get_coords_list_chord_relative(
                max_airfoil_points=max_airfoil_points, curvature_exp=curvature_exp)
            save_data(transformation_kwargs, os.path.join(blade_file_dir, "transformation.json"))

        # Set the default grid bounds value
        if grid_bounds is None:
            grid_bounds = [-5.0, 5.0, -5.0, 5.0]

        # Write the header (line 1: airfoil name, line 2: grid bounds values separated by spaces)
        header = airfoil_sys_name + "\n" + " ".join([str(gb) for gb in grid_bounds])

        # Determine the correct ordering for the airfoils. MSES expects airfoils to be ordered from top to bottom
        max_y = [np.max(coords[:, 1]) for coords in mea_coords_list]
        airfoil_order = np.argsort(max_y)[::-1]

        # Loop through the airfoils in the correct order
        mea_coords = None
        for airfoil_idx in airfoil_order:
            airfoil_coords = mea_coords_list[airfoil_idx]  # Extract the airfoil coordinates for this airfoil
            if mea_coords is None:
                mea_coords = airfoil_coords
            else:
                mea_coords = np.row_stack((mea_coords, np.array([999.0, 999.0])))  # MSES-specific airfoil delimiter
                mea_coords = np.row_stack((mea_coords, airfoil_coords))  # Append this airfoil's coordinates to the mat.

        # Generate the full file path
        blade_file_path = os.path.join(blade_file_dir, f"blade.{airfoil_sys_name}")

        # Save the coordinates to file
        np.savetxt(blade_file_path, mea_coords, header=header, comments="")

        # Get the airfoil name order
        airfoil_name_order = [airfoil.name() for airfoil in [self.airfoils[idx] for idx in airfoil_order]]

        return blade_file_path, airfoil_name_order

    def write_to_IGES(self, file_name: str):
        """
        Writes the airfoil system to file using the IGES file format.

        Parameters
        ==========
        file_name: str
            Path to IGES file
        """
        bez_IGES_entities = [
            [BezierIGES(np.column_stack((c.P[:, 0], np.zeros(len(c.P)), c.P[:, 1]))) for c in a.curve_list]
            for a in self.airfoils]
        entities_flattened = list(itertools.chain.from_iterable(bez_IGES_entities))
        iges_generator = IGESGenerator(entities_flattened)
        iges_generator.generate(file_name)

    def get_max_x_extent(self):
        coords_list, _ = self.get_coords_list_chord_relative()
        max_x_list = [max(coords[:, 0]) for coords in coords_list]
        return max(max_x_list)

    def get_dict_rep(self):
        return {"airfoils": [a.name() for a in self.airfoils]}



import os
import typing

import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import pyqtSignal

from pymead import q_settings, GUI_SETTINGS_DIR
from pymead.core.bezier import Bezier
from pymead.core.line import LineSegment, PolyLine
from pymead.core.parametric_curve import PCurveData, ParametricCurve
from pymead.core.point import PointSequence, Point
from pymead.gui.draggable_point import DraggablePoint
from pymead.utils.misc import convert_str_to_Qt_dash_pattern
from pymead.utils.read_write_files import load_data

q_settings_descriptions = load_data(os.path.join(GUI_SETTINGS_DIR, "q_settings_descriptions.json"))


class HoverableCurve(pg.PlotCurveItem):
    """
    Hoverable curve item to be drawn on an AirfoilCanvas. Inspired in part by https://stackoverflow.com/a/68857695.
    """

    sigCurveHovered = pyqtSignal(object, object)
    sigCurveNotHovered = pyqtSignal(object, object)
    sigRemove = pyqtSignal(object)
    sigBezierAdded = pyqtSignal(object)
    sigLineAdded = pyqtSignal(object)
    sigPolyLineAdded = pyqtSignal(object)
    sigLineItemAdded = pyqtSignal(object)

    def __init__(self, curve_type: str, *args, **kwargs):
        """
        Parameters
        ==========
        curve_type: str
            Type of curve to be drawn

        args
            Positional arguments which are passed to ``pyqtgraph.PlotCurveItem``

        kwargs
            Keyword arguments which are passed to ``pyqtgraph.PlotCurveItem``
        """
        implemented_curve_types = ["Bezier", "LineSegment", "PolyLine"]
        if curve_type not in implemented_curve_types:
            raise NotImplementedError(f"Curve type {curve_type} either incorrectly labeled or not yet implemented."
                                      f"Currently implemented curve types: {implemented_curve_types}.")
        super().__init__(*args, **kwargs)
        self.hoverable = True
        self.setAcceptHoverEvents(True)
        self.curve_type = curve_type
        self.point_items = []
        self.control_point_line_items = None
        self.parametric_curve = None

    def hoverEvent(self, ev):
        """
        Trigger custom signals when a hover event is detected. Only active when ``hoverable==True``.
        """
        if self.hoverable:
            if hasattr(ev, "_scenePos") and self.mouseShape().contains(ev.pos()):
                self.sigCurveHovered.emit(self, ev)
            else:
                self.sigCurveNotHovered.emit(self, ev)

    def updateCurveData(self, data: np.ndarray):
        """
        Updates the data in the curve from a ``numpy`` array.

        Parameters
        ==========
        data: np.ndarray
            Two-column array, where the two columns represent the :math:`x` and :math:`y` values of a sequence of points
            that describes the curve. In the case of splines, these points are the control points.

        Returns
        =======
        PCurveData
            Parameter vector and "xy" ``PointSequence`` describing the parametric curve
        """
        point_sequence = PointSequence.generate_from_array(data)
        if self.curve_type == "Bezier":
            if self.parametric_curve is None:
                self.parametric_curve = Bezier(point_sequence)
                self.sigBezierAdded.emit(self.parametric_curve)
            else:
                self.parametric_curve.set_point_sequence(point_sequence)
        elif self.curve_type == "LineSegment":
            if self.parametric_curve is None:
                self.parametric_curve = LineSegment(point_sequence)
                self.sigLineAdded.emit(self.parametric_curve)
            else:
                self.parametric_curve.set_point_sequence(point_sequence)
        elif self.curve_type == "PolyLine":
            if self.parametric_curve is None:
                self.parametric_curve = PolyLine(point_sequence)
                self.sigPolyLineAdded.emit(self.parametric_curve)
            else:
                self.parametric_curve.set_point_sequence(point_sequence)
        curve_data = self.parametric_curve.evaluate()
        return curve_data

    def getControlPointArray(self):
        control_point_array = np.array([item.data["pos"][0, :] for item in self.point_items])
        return control_point_array

    def updateCurveItem(self, curve_data: PCurveData or np.ndarray or None = None):
        """
        Function that wraps around ``updateCurveData`` and implements other input data types.

        Parameters
        ==========
        curve_data: PCurveData or np.ndarray or None
            If ``PCurveData``, directly update the GUI implementation of the curve. If an array, first update the curve
            data and then update the GUI. If ``None``, first need to extract an array of points from the ``point_items``
            attribute.
        """
        if curve_data is None:
            if len(self.point_items) == 0:
                raise ValueError("No curve data specified and no point items set in Curve Item")
            curve_data = self.getControlPointArray()
        if isinstance(curve_data, np.ndarray):
            curve_data = self.updateCurveData(curve_data)
        self.setData(x=curve_data.xy[:, 0], y=curve_data.xy[:, 1])
        # self.updateControlPointNetItem()

    def updateCanvasItem(self, curve_data: PCurveData):
        arr = curve_data.xy
        self.setData(x=arr[:, 0], y=arr[:, 1])

    @staticmethod
    def generateControlPointNetLine(parametric_curve: ParametricCurve, point_items: typing.List[DraggablePoint]):
        line_item = HoverableCurve(curve_type="LineSegment")
        line_item.parametric_curve = parametric_curve
        line_item.point_items = point_items
        line_item.updateCurveItem()
        line_item.hoverable = True
        # line_item.setZValue(-100)
        return line_item

    def generateControlPointNetItems(self):
        if self.curve_type != "Bezier":
            return
        self.control_point_line_items = []
        for line_idx, line in enumerate(self.parametric_curve.control_point_lines):
            point_items = [self.point_items[line_idx], self.point_items[line_idx + 1]]
            line_item = self.generateControlPointNetLine(line, point_items)
            self.control_point_line_items.append(line_item)
            self.sigLineItemAdded.emit(line_item)
        # Make the control point net behind everything else
        # self.control_point_net.setZValue(-100)

    def removeControlPointFromNet(self, pt: Point):
        if self.curve_type != "Bezier":
            return

        idx = self.point_items.index(pt)

        n_ctrl_pts = len(self.parametric_curve.point_sequence())

        # Case: first point
        if idx == 0:
            self.control_point_line_items.pop(0)

        # Case: last point
        elif idx == n_ctrl_pts - 1:
            self.control_point_line_items.pop(-1)

        # General case:
        else:
            # First, generate a line spanning the points on either side of the point we are deleting
            point_sequence = self.parametric_curve.point_sequence().extract_subsequence(indices=[idx - 1, idx + 1])
            line = LineSegment(point_sequence=point_sequence, reference=True)
            new_line_item = self.generateControlPointNetLine(
                parametric_curve=line, point_items=[self.point_items[idx - 1], self.point_items[idx + 1]]
            )

            # Then, delete the lines connecting to the points we are deleting
            self.control_point_line_items.pop(idx)
            self.control_point_line_items.pop(idx - 1)

            self.control_point_line_items.insert(idx, new_line_item)
            self.sigLineItemAdded.emit(new_line_item)

        # TODO: add a new line if necessary! only true for the general case

    # def updateControlPointNetItem(self):
    #     if self.control_point_line_items is None:
    #         return
    #     for line in self.control_point_line_items:
    #         line.setData(x=line.parametric_curve)

    def remove(self):
        for point in self.point_items:
            point.curves.remove(self)

    def setCurveStyle(self, mode: str = "default"):
        implemented_style_modes = ["default", "hovered"]
        if mode not in implemented_style_modes:
            raise NotImplementedError(f"Selected mode ({mode}) not implemented. "
                                      f"Currently implemented modes: {implemented_style_modes}.")
        self.setPen(pg.mkPen(color=q_settings.value(f"curve_{mode}_pen_color",
                                                    q_settings_descriptions[f"curve_{mode}_pen_color"][1]),
                             width=q_settings.value(f"curve_{mode}_pen_width",
                                                    q_settings_descriptions[f"curve_{mode}_pen_width"][1])))

        if self.parametric_curve is not None and self.parametric_curve.reference:
            dash_str = q_settings.value(f"cpnet_default_pen_dash", q_settings_descriptions[f"cpnet_default_pen_dash"][1])
            dash = convert_str_to_Qt_dash_pattern(dash_str)
            self.setPen(style=dash)

    def setControlPointNetStyle(self, mode: str = "default"):
        # if self.control_point_net is None:
        #     return
        # implemented_style_modes = ["default"]
        # if mode not in implemented_style_modes:
        #     raise NotImplementedError(f"Selected mode ({mode}) not implemented. "
        #                               f"Currently implemented modes: {implemented_style_modes}.")
        #
        # dash_str = q_settings.value(f"cpnet_{mode}_pen_dash", q_settings_descriptions[f"cpnet_{mode}_pen_dash"][1])
        # dash = convert_str_to_Qt_dash_pattern(dash_str)
        #
        # self.control_point_net.setPen(pg.mkPen(
        #     color=q_settings.value(f"cpnet_{mode}_pen_color",
        #                            q_settings_descriptions[f"cpnet_{mode}_pen_color"][1]),
        #     width=q_settings.value(f"cpnet_{mode}_pen_width",
        #                            q_settings_descriptions[f"cpnet_{mode}_pen_width"][1]),
        #     style=dash))
        pass



import pyqtgraph as pg


class DragGraph:
    def __init__(self, pen=None, size: tuple = (1000, 300), background_color: str = 'w'):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)
        # self.w.setWindowTitle('Airfoil')
        self.w.setBackground(background_color)
        self.v = self.w.addPlot(pen=pen)
        self.v.setLabel(axis='bottom', text='Generation')
        self.v.setLabel(axis='left', text='Drag (Counts)')
        self.v.showGrid(x=True, y=True)
        self.legend = self.v.addLegend(offset=(30, 30))
        self.pg_plot_handle_Cd = self.v.plot(pen=pg.mkPen(color='indianred'), name='Cd', symbol='s', symbolBrush=pg.mkBrush('indianred'))
        self.pg_plot_handle_Cdp = self.v.plot(pen=pg.mkPen(color='gold'), name='Cdp', symbol='o', symbolBrush=pg.mkBrush('gold'))
        self.pg_plot_handle_Cdf = self.v.plot(pen=pg.mkPen(color='limegreen'), name='Cdf', symbol='t', symbolBrush=pg.mkBrush('limegreen'))
        self.pg_plot_handle_Cdv = self.v.plot(pen=pg.mkPen(color='mediumturquoise'), name='Cdv', symbol='d', symbolBrush=pg.mkBrush('mediumturquoise'))
        self.pg_plot_handle_Cdw = self.v.plot(pen=pg.mkPen(color='violet'), name='Cdw', symbol='x', symbolBrush=pg.mkBrush('violet'))

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)


class CpGraph:
    def __init__(self, pen=None, size: tuple = (1000, 300), background_color: str = 'w'):
        pg.setConfigOptions(antialias=True)

        if pen is None:
            pen = pg.mkPen(color='cornflowerblue', width=2)

        self.w = pg.GraphicsLayoutWidget(show=True, size=size)
        # self.w.setWindowTitle('Airfoil')
        self.w.setBackground(background_color)
        self.v = self.w.addPlot(pen=pen)
        self.v.setLabel(axis='bottom', text='x')
        self.v.setLabel(axis='left', text='Cp')
        self.v.invertY(True)
        self.v.showGrid(x=True, y=True)

    def set_background(self, background_color: str):
        self.w.setBackground(background_color)



import os
import itertools

import matplotlib.legend
import matplotlib.lines
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches


def format_axis_scientific(ax: plt.Axes, font_family: str = "serif", font_size: int = 16):
    """
    Formats a Matplotlib ``Axes`` similar to the IEEE style
    """
    for tick in ax.get_xticklabels():
        tick.set_fontfamily(font_family)
        tick.set_fontsize(font_size)
    for tick in ax.get_yticklabels():
        tick.set_fontfamily(font_family)
        tick.set_fontsize(font_size)
    ax.minorticks_on()
    ax.tick_params(which="minor", direction="in", left=True, bottom=True, top=True, right=True)
    ax.tick_params(which="major", direction="in", left=True, bottom=True, top=True, right=True)


def show_save_fig(fig: plt.Figure, save_base_dir: str, file_name_stub: str, show: bool = False, save: bool = True,
                  show_first: bool = True, save_ext: tuple = (".png", ".svg", ".pdf")):

    def show_plot():
        plt.show()

    def save_plot():
        for ext in save_ext:
            save_kwargs = {}
            if ext in [".png", ".jpg"]:
                save_kwargs["dpi"] = 600
            fig.savefig(os.path.join(save_base_dir, file_name_stub + ext), bbox_inches="tight", **save_kwargs)

    if show_first:
        if show:
            show_plot()
        if save:
            save_plot()
    else:
        if save:
            save_plot()
        if show:
            show_plot()


def dark_mode(fig: plt.Figure):
    """
    Sets a dark theme for many elements of a Matplotlib ``Figure``

    Parameters
    ==========
    fig: plt.Figure
        Figure to modify
    """

    def _format_axis(axis):
        axis.set_facecolor(black)
        for spine in axis.spines.values():
            spine.set_color("w")
        axis.xaxis.label.set_color("w")
        axis.yaxis.label.set_color("w")
        axis.tick_params(axis="x", colors="w")
        axis.tick_params(axis="y", colors="w")
        axis.set_title(axis.get_title(), color="w")

    def _format_legend(lg):
        for text in lg.get_texts():
            text.set_color("w")
        frame = lg.get_frame()
        frame.set_facecolor("k")

    ax_list = fig.get_axes()
    black = "#121212"
    for ax in ax_list:

        _format_axis(ax)

        for child in ax.get_children():
            print(f"{child = }")
            if isinstance(child, plt.Annotation):
                if child.get_color() == "black":
                    child.set_color("w")
            if isinstance(child, plt.Axes):
                _format_axis(child)
                for grandchild in child.get_children():
                    print(f"{grandchild = }")
                    if isinstance(grandchild, plt.Text):
                        print(f"{grandchild.get_color() = }")
                        if grandchild.get_color() == "black":
                            grandchild.set_color("w")
            if isinstance(child, matplotlib.legend.Legend):
                _format_legend(child)
            if isinstance(child, mpatches.ConnectionPatch):
                child.set_color("w")
            if isinstance(child, mpatches.Rectangle):
                child.set_edgecolor("w")

    fig.set_facecolor(black)
    for leg in fig.legends:
        _format_legend(leg)


def legend_entry_flip(items, ncol):
    """Solution for legend entry flipping from StackOverflow user Avaris @ https://stackoverflow.com/a/10101532"""
    return itertools.chain(*[items[i::ncol] for i in range(ncol)])




import os
import re
from copy import deepcopy

import numpy as np
from pymoo.core.problem import Problem
from pymoo.factory import get_decomposition
from pymoo.util.display import Display
from pymoo.util.termination.default import MultiObjectiveDefaultTermination
from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTermination

multi_point_keys_mses = {
    0: 'MACHIN',
    1: 'REYNIN',
    2: 'ALFAIN',
    3: 'CLIFIN',
    4: 'P',
    5: 'T',
    6: 'L',
    7: 'R',
    8: 'rho',
    9: 'gam',
    10: 'ACRIT',
    11: 'MCRIT',
    12: 'MUCON',
    13: 'XTRSupper',
    14: 'XTRSlower',
    15: 'ISDELH',
    16: 'XCDELH',
    17: 'PTRHIN',
    18: 'ETAH',
    19: 'ISMOM',
    20: 'IFFBC',
}

multi_point_keys_xfoil = {
    0: "Ma",
    1: "Re",
    2: "alfa",
    3: "Cl",
    4: "CLI",
    5: "P",
    6: "T",
    7: "L",
    8: "R",
    9: "rho",
    10: "gam",
    11: "N",
    12: "xtr_upper",
    13: "xtr_lower",
}


def read_stencil_from_array(data: np.ndarray, tool: str):
    """Read the Multi-Point stencil from a text file and converts it to a list of dictionaries"""
    if tool == "MSES":
        multi_point_keys = multi_point_keys_mses
    elif tool == "XFOIL":
        multi_point_keys = multi_point_keys_xfoil
    else:
        raise ValueError("Either 'MSES' or 'XFOIL' must be selected for the tool in the multipoint stencil")

    if data.ndim == 1:
        return [{"variable": multi_point_keys[int(data[0])], "index": int(data[1]), "points": data[2:].tolist()}]
    elif data.ndim == 2:
        return [{'variable': multi_point_keys[int(col[0])], 'index': int(col[1]),
                 'points': col[2:].tolist()}
                for col in data.T]
    else:
        raise ValueError(f"Discovered multipoint data that was not 1-D or 2-D. {data.ndim = }")


def termination_condition(param_dict: dict):
    """
    Termination criteria for the optimization.

    Parameters
    ==========
    param_dict: dict
        Parameter dictionary for the optimization

    pymoo.util.termination.default.MultiObjectiveDefaultTermination
        Termination object used to define convergence for the genetic algorithm
    """
    termination = MultiObjectiveDefaultTermination(
        x_tol=param_dict['x_tol'],
        cv_tol=param_dict['cv_tol'],
        f_tol=param_dict['f_tol'],
        nth_gen=param_dict['nth_gen'],
        n_last=param_dict['n_last'],
        n_max_gen=param_dict['n_max_gen'],
        n_max_evals=param_dict['n_max_evals']
    )
    return termination


def calculate_warm_start_index(warm_start_generation: int, warm_start_directory: str):
    """
    Calculates the generation from which to restart the optimization. If the specified warm start generation is not
    found in the ``algorithm_gen_xxx.pkl`` files, an error is raised.

    Parameters
    ==========
    warm_start_generation: int
        Generation from which to restart the optimization. If this value is ``-1``, the optimization will start
        from the most recent point.

    warm_start_directory: str
        Path to the optimization directory containing the ``algorithm_gen_xxx.pkl`` files

    Returns
    =======
    int
        The index of the generation to start from.
    """
    generations = []
    for root, _, files in os.walk(warm_start_directory):
        for idx, f in enumerate(files):
            file_without_ext = os.path.splitext(f)[0]
            if re.split('_', file_without_ext)[0] == 'algorithm':
                idx = re.split('_', file_without_ext)[-1]
                generations.append(int(idx))
    generations = sorted(generations)
    if warm_start_generation not in generations and warm_start_generation != -1:
        raise ValueError(f'Invalid warm start generation. A value of {warm_start_generation} was input, and the valid '
                         f'generations in the directory are {generations}')
    if len(generations) > 0:
        warm_start_index = generations[warm_start_generation]
    else:
        warm_start_index = 0

    return warm_start_index


class TPAIOPT(Problem):
    def __init__(self, n_var: int, n_obj: int, n_constr: int, xl: int or list or np.ndarray,
                 xu: int or list or np.ndarray, param_dict: dict):
        """
        Simple problem statement for the ``pymoo`` package.

        Parameters
        ==========
        n_var: int
            Number of design variables (equal to the length of the normalized paramter list)

        n_obj: int
            Number of objectives

        n_constr: int
            Number of constraints

        xl: int or list or np.ndarray
            Lower bounds for the parameters. If ``int``, all lower bounds are equal.

        xu: int or list or np.ndarray
            Upper bounds for the parameters. If ``int``, all lower bounds are equal.

        param_dict: dict
            Parameter dictionary used for the shape optimization.
        """
        super().__init__(n_var=n_var, n_obj=n_obj, n_constr=n_constr, xl=xl, xu=xu)

        self.param_dict = deepcopy(param_dict)

    def _evaluate(self, X, out, *args, **kwargs):
        pass


class CustomDisplay(Display):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.term = MultiObjectiveSpaceToleranceTermination()
        self.progress_dict = None

    def set_progress_dict(self, progress_dict: dict):
        self.progress_dict = progress_dict

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        F = algorithm.pop.get("F")
        G = algorithm.pop.get("G")
        # CV = algorithm.pop.get("CV")
        weights = np.array([0.5, 0.5])
        decomp = get_decomposition("asf")
        I = decomp.do(F, weights).argmin()
        n_nds = len(algorithm.opt)
        f_best = [F[I][i] for i in range(F.shape[1])]
        f_min = [F[:, i].min() for i in range(F.shape[1])]
        f_mean = [np.mean(F[:, i]) for i in range(F.shape[1])]
        g_best, g_min, g_mean = None, None, None
        if G[0] is not None:
            g_best = [G[I][i] for i in range(G.shape[1])]
            g_min = [G[:, i].min() for i in range(G.shape[1])]
            # cv_min = CV[:, 0].min()
            g_mean = [np.mean(G[:, i]) for i in range(G.shape[1])]
        self.output.append("n_nds", n_nds, width=7)
        self.term.do_continue(algorithm)

        max_from, eps = "-", "-"

        if len(self.term.metrics) > 0:
            metric = self.term.metrics[-1]
            tol = self.term.tol
            delta_ideal, delta_nadir, delta_f = metric["delta_ideal"], metric["delta_nadir"], metric["delta_f"]

            if delta_ideal > tol:
                max_from = "ideal"
                eps = delta_ideal
            elif delta_nadir > tol:
                max_from = "nadir"
                eps = delta_nadir
            else:
                max_from = "f"
                eps = delta_f

        self.output.append("eps", eps)
        self.output.append("indicator", max_from)
        for i, f in enumerate(f_best):
            self.output.append(f"f{i + 1}_best", f)
        for i, f in enumerate(f_min):
            self.output.append(f"f{i + 1}_min", f)
        for i, f in enumerate(f_mean):
            self.output.append(f"f{i + 1}_mean", f)

        if G[0] is not None:
            for i, g in enumerate(g_best):
                self.output.append(f"g{i + 1}_best", g)
            for i, g in enumerate(g_min):
                self.output.append(f"g{i + 1}_min", g)
            for i, g in enumerate(g_mean):
                self.output.append(f"g{i + 1}_mean", g)

        self.set_progress_dict(self.output.attrs)



from PyQt5.QtWidgets import QDialog, QLineEdit, QVBoxLayout, QDialogButtonBox, QTreeWidgetItem


class RenamePopup(QDialog):
    def __init__(self, name: str, item_to_rename, parent=None):
        super().__init__(parent)
        self.item_to_rename = item_to_rename
        self.name = name
        self.setWindowTitle("Rename")

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel

        self.button_box = QDialogButtonBox(QBtn)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.layout = QVBoxLayout()
        self.q_line_edit = QLineEdit(name)
        self.layout.addWidget(self.q_line_edit)
        self.layout.addWidget(self.button_box)
        self.setLayout(self.layout)

    def accept(self) -> None:
        self.item_to_rename.setText(0, self.q_line_edit.text())
        self.done(0)

    def reject(self) -> None:
        self.done(0)



import numpy as np

from pymead.plugins.IGES.curves import BezierIGES
from pymead.plugins.IGES.iges_generator import IGESGenerator


def main():
    file_name = "iges_generation_example.igs"
    P = np.array([
        [0., 0., 0.],
        [100., 50., 80.],
        [-60., 30., -50.],
        [220., -80., 125.],
        [20., 20., 20.]
    ])
    bez = BezierIGES(P)
    iges_generator = IGESGenerator(entities=[bez])
    iges_generator.generate(file_name)


if __name__ == "__main__":
    main()



from functools import partial

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGridLayout, QCheckBox, QLabel, QWidget

from pymead.gui.dialogs import PymeadDialog


class ShowHideDialog(PymeadDialog):
    def __init__(self, parent, state: dict, theme: dict):
        widget = QWidget()
        super().__init__(parent=parent, window_title="Show/Hide", widget=widget, theme=theme)
        self.lay = QGridLayout()
        widget.setLayout(self.lay)
        self.state = state
        self.addRows()
        self.lay.setColumnStretch(0, 0)
        self.lay.setColumnStretch(1, 1)
        self.setMinimumWidth(250)

    def addRows(self):

        items = {
            "points": "Points",
            "lines": "Lines",
            "bezier": "Bezier",
            "airfoils": "Airfoils",
            "geocon": "Constraints"
        }

        for row_idx, (sub_container, title) in enumerate(items.items()):
            checkbox = QCheckBox(self)
            checkbox.setChecked(self.state[sub_container])
            checkbox.clicked.connect(partial(self.parent().showHidePymeadObjs, sub_container))
            checkbox.clicked.connect(partial(self.onCheckBoxStateChanged, sub_container))

            label = QLabel(title, self)
            self.lay.addWidget(checkbox, row_idx, 0)
            self.lay.addWidget(label, row_idx, 1)

    def onCheckBoxStateChanged(self, sub_container: str, shown: bool):
        self.state[sub_container] = shown



import json
import os
import pickle
import typing

from PyQt5.QtCore import QStandardPaths

from pymead import q_settings


def save_data(var, file):
    if os.path.splitext(file)[-1] == '.pkl':
        with open(file, 'wb') as file:
            pickle.dump(var, file, protocol=-1)
    elif os.path.splitext(file)[-1] in ['.json', '.jmea']:
        with open(file, 'w') as file:
            json.dump(var, file, indent=4)
    else:
        raise Exception('Invalid file extension for data save! Current available choices: .pkl, .json, .jmea')


def load_data(file):
    if os.path.splitext(file)[-1] == '.pkl':
        with open(file, 'rb') as file:
            var = pickle.load(file)
        return var
    elif os.path.splitext(file)[-1] in ['.json', '.jmea']:
        with open(file, 'r') as file:
            var = json.load(file)
        return var
    else:
        raise Exception('Invalid file extension for data load! Current available choices: .pkl, .json, .jmea')


def write_tuple_tuple_to_file(fname: str, data: typing.Tuple[tuple]):
    """Data must be 2-D"""
    with open(fname, 'w') as f:
        for row in data:
            for col in row:
                f.write(f"{col} ")
            f.write("\n")


def load_documents_path(settings_var: str):
    """
    Utility function that returns a specified QSettings path location if saved, otherwise returns the Documents location

    Parameters
    ==========
    settings_var: str
        Key to the path location in ``pymead.q_settings``

    Returns
    =======
    str
        Path specified by ``settings_var``, otherwise the Documents location
    """
    if q_settings.contains(settings_var):
        path = q_settings.value(settings_var)
    else:
        path = QStandardPaths.writableLocation(QStandardPaths.DocumentsLocation)
    return path



import os
import re
import typing

import pandas as pd
import numpy as np


flow_var_idx = {"M": 7, "Cp": 8, "p": 3, "rho": 2, "u": 4, "v": 5, "V": 6, "Cpt": 9, "dCpt": 10, "dCp": 11}


def read_Cp_from_file_xfoil(fname: str):
    """
    Reads the :math:`C_p` data from the format output by XFOIL and converts it to a *numpy* array

    Parameters
    ==========
    fname: str
      File from which to read the :math:`C_p` data

    Returns
    =======
    dict
      Dictionary containing 1-D *numpy* arrays for :math:`x`, :math:`y`, and :math:`C_p`
    """
    y_in_header = False
    header_line_idx = None
    with open(fname, "r") as f:
        for idx, line in enumerate(f):
            if "Cp" not in line:
                continue
            header_line_idx = idx
            if "y" in line:
                y_in_header = True
            break
        else:
            raise ValueError("Could not detect the header line in the XFOIL Cp file")

    # Read in the Cp data for XFOIL versions 6.93 or 6.99 (other versions are untested)
    names = ["x", "y", "Cp"] if y_in_header else ["x", "Cp"]
    cols = {"x": 0, "y": 1, "Cp": 2} if y_in_header else {"x": 0, "Cp": 1}
    df = pd.read_csv(fname, skiprows=header_line_idx + 1, names=names, sep=r"\s+", engine='python')
    array_ = df.to_numpy()
    return {name: array_[:, cols[name]] for name in names}


def read_aero_data_from_xfoil(fpath: str, aero_data: dict):
    """
    Reads aerodynamic data from XFOIL's output log

    Parameters
    ==========
    fpath: str
      File path to the XFOIL log

    aero_data: dict
      Dictionary in which the aerodynamic data will be written

    Returns
    =======
    str, str
      The first and second lines of the XFOIL log files containing aerodynamic output data
    """
    line1 = None
    line2 = None
    for line in read_reverse_order(fpath):
        if 'VISCAL:  Convergence failed' in line:
            aero_data['converged'] = False
            break
        elif 'Cm' in line:
            line1 = line
        elif 'CL' in line:
            line2 = line
            aero_data['converged'] = True
            break
    else:
        aero_data['converged'] = False
        aero_data['errored_out'] = True

    return line1, line2


def read_reverse_order(file_name):
    """
    Reads a file in reverse order. From `<https://thispointer.com/python-read-a-file-in-reverse-order-line-by-line/>`_
    """
    # Open file for reading in binary mode
    with open(file_name, 'rb') as read_obj:
        # Move the cursor to the end of the file
        read_obj.seek(0, os.SEEK_END)
        # Get the current position of pointer i.e eof
        pointer_location = read_obj.tell()
        # Create a buffer to keep the last read line
        buffer = bytearray()
        # Loop till pointer reaches the top of the file
        while pointer_location >= 0:
            # Move the file pointer to the location pointed by pointer_location
            read_obj.seek(pointer_location)
            # Shift pointer location by -1
            pointer_location = pointer_location - 1
            # read that byte / character
            new_byte = read_obj.read(1)
            # If the read byte is new line character then it means one line is read
            if new_byte == b'\n':
                # Fetch the line from buffer and yield it
                yield buffer.decode()[::-1]
                # Reinitialize the byte array to save next line
                buffer = bytearray()
            else:
                # If last read character is not eol then add it in buffer
                buffer.extend(new_byte)
        # As file is read completely, if there is still data in buffer, then its the first line.
        if len(buffer) > 0:
            # Yield the first line too
            yield buffer.decode()[::-1]


def read_forces_from_mses(search_file: str):
    """
    This function uses regex to extract the aerodynamic force coefficients of the airfoil [system] from the
    ``list forces`` option available in the MPLOT menu

    Parameters
    ==========
    search_file: str
      "Forces" file to be read (log of the MPLOT function with ``mode=="forces"``)

    Returns
    =======
    dict
      Dictionary containing the aerodynamic force coefficients
    """
    CL_lines = []
    alpha_line, vw_line, fp_line, ad_line = None, None, None, None
    with open(search_file, 'r') as f:
        file_out = f.readlines()
    for line in file_out:
        if re.search("CL", line):
            CL_lines.append(line)
        if 'alpha' in line:
            alpha_line = line
        if 'viscous' in line:
            vw_line = line
        if 'friction' in line:
            fp_line = line
        if 'CDh' in line:
            ad_line = line
    try:
        required_line = CL_lines[1]  # line we need happens the second time the string "CL" is mentioned
        split_line = required_line.split()  # split the up the line to grab the actual numbers
        forces = {  # write as a dictionary
            'Cl': float(split_line[2]),  # CL = VALUE
            'Cd': float(split_line[5]),  # CD = VALUE
            'Cm': float(split_line[8])  # CM = VALUE
        }

        # Find the angle of attack
        alpha_line = alpha_line.split('=')
        alpha_line = alpha_line[-1].split()
        forces['alf'] = float(alpha_line[-2])

        # Find the viscous drag and wave drag coefficients in the drag breakdown (these two values + Cdh should add to Cd)
        vw_line = vw_line.split('=')
        forces['Cdv'] = float(vw_line[-2].split()[0])
        forces['Cdw'] = float(vw_line[-1].split()[0])

        # Find the friction drag and pressure drag coefficients in the drag breakdown (these two values should add to Cd)
        fp_line = fp_line.split('=')
        forces['Cdf'] = float(fp_line[-2].split()[0])
        forces['Cdp'] = float(fp_line[-1].split()[0])

        # Find the "drag" due to the actuator disk:
        if ad_line is not None:
            ad_line = ad_line.split('=')
            forces['Cdh'] = float(ad_line[-1].split()[0])
        else:
            forces['Cdh'] = 0.0
    except:
        forces = {'Cl': 0.0, 'Cd': 1000.0, 'Cm': 1000.0, 'alf': 0.0, 'Cdv': 1000.0, 'Cdw': 1000.0,
                  'Cdf': 1000.0, 'Cdp': 1000.0, 'Cdh': 1000.0, 'CPK': 1000.0}

    # print(f"{forces = }")
    return forces


def read_grid_stats_from_mses(src_file: str):
    """
    Reads grid statistics from an MPLOT output file and outputs them to a dictionary

    Parameters
    ==========
    src_file: str
      The file from which to read the grid statistics

    Returns
    =======
    dict
      The grid statistics in a Python dictionary format
    """
    with open(src_file, 'r') as f:
        lines = f.readlines()

    grid_stats = {'grid_size': None, 'numel': None, 'Jside1': [], 'Jside2': [],
                  'ILE1': [], 'ILE2': [], 'ITE1': [], 'ITE2': [], }

    for line in lines:
        if 'Grid size' in line:
            line_split = line.split('=')[-1]
            # print(f"{[idx for idx in line_split.split()] = }")
            grid_stats['grid_size'] = [int(idx) for idx in line_split.split()]
        elif 'Number' in line:
            line_split = line.split(':')[-1]
            grid_stats['numel'] = int(line_split.strip())
        elif 'Jside1' in line:
            line_split = line.split('=')
            grid_stats['Jside1'].append(int(line_split[1].split()[0]))
            grid_stats['ILE1'].append(int(line_split[2].split()[0]))
            grid_stats['ITE1'].append(int(line_split[3].split()[0]))
        elif 'Jside2' in line:
            line_split = line.split('=')
            grid_stats['Jside2'].append(int(line_split[1].split()[0]))
            grid_stats['ILE2'].append(int(line_split[2].split()[0]))
            grid_stats['ITE2'].append(int(line_split[3].split()[0]))

    return grid_stats


def read_actuator_disk_data_mses(mses_log_file: str, grid_stats: dict) -> typing.List[dict]:
    """
    Reads actuator disk data from the MSES log file (usually ``mses.log``).

    Parameters
    ==========
    mses_log_file: str
        Log file generated by MSES after CFD evaluation, usually ``mses.log``.

    grid_stats: dict
        Grid statistics dictionary generated by ``read_grid_stats_from_mses`` in this module.

    Returns
    =======
    typing.List[dict]
        List of actuator disk data, with each element in the list representing a different actuator disk
    """
    with open(mses_log_file, "r") as f:
        lines = f.readlines()

    lines_AD = []
    for line_idx, line in enumerate(lines):
        if "Actuator disk plane" in line:
            lines_AD.append(line_idx)

    data_AD = []
    for line_idx in lines_AD:
        line1, line2, line3 = lines[line_idx].split(), lines[line_idx + 1].split(), lines[line_idx + 2].split()
        data_AD.append({"x_c": float(line1[6]),
                        "i": int(line1[9]),
                        "j_start": int(line1[12]),
                        "j_end": int(line1[14]),
                        "pt2_pt1": float(line2[2]),
                        "dCpo": float(line2[5]),
                        "ht2_ht1": float(line3[2]),
                        "etah": float(line3[5])
                        })

    for data in data_AD:
        flow_section_idx = None
        if data["j_start"] == 1:
            flow_section_idx = 0
        else:
            for j_side_idx, j_side in enumerate(grid_stats["Jside1"][::-1]):
                if data["j_start"] == j_side:
                    flow_section_idx = j_side_idx + 1
        data["flow_section_idx"] = flow_section_idx
        data["field_i_up"] = data["i"] - 2
        data["field_i_down"] = data["i"] - 1
        data["field_j_start"] = data["j_start"] - flow_section_idx - 1
        data["field_j_end"] = data["j_end"] - flow_section_idx - 1

    return data_AD


def read_bl_data_from_mses(src_file: str) -> typing.List[dict]:
    r"""
    Reads boundary layer information from ``bl.*`` files generated by MSES

    Parameters
    ==========
    src_file:
      Boundary layer file, usually with the filename ``bl.{airfoil_name}``

    Returns
    =======
    typing.List[dict]
      Boundary layer data list (length :math:`M`) of dictionaries, where :math:`M` is the number of airfoil sides
      (2 times the number of airfoils)
    """
    with open(src_file, 'r') as f:
        line1 = f.readline()
    header_line = line1.replace(' #', '').split()
    header_line.pop(0)
    header_line[-1] = 'Pend'
    df = pd.read_csv(src_file, sep=r"\s+", skiprows=2, names=header_line)
    bl = [{}]
    side_idx = 0
    # old_x_val = 0
    for idx, s_val in enumerate(df.s):
        if idx == 0:
            for var_name in df.columns:
                bl[side_idx][var_name] = [getattr(df, var_name)[idx]]
        else:
            if s_val == 0.0:  # Criterion for the start of a new airfoil side: arc length = 0
                side_idx += 1
                bl.append({})
                for var_name in df.columns:
                    bl[side_idx][var_name] = [getattr(df, var_name)[idx]]
            else:
                for var_name in df.columns:
                    bl[side_idx][var_name].append(getattr(df, var_name)[idx])
    return bl


def read_Mach_from_mses_file(mses_file: str) -> float:
    """
    Reads the freestream Mach number from the MSES settings file (usually of the form ``mses.*``).

    Parameters
    ==========
    mses_file: str
        Settings file for MSES, most likely of the form ``mses.*``

    Returns
    =======
    float
        Freestream Mach number
    """
    with open(mses_file, "r") as f:
        lines = f.readlines()
    mach_line = lines[2]
    mach_split = mach_line.split()
    mach_float = float(mach_split[0])
    return mach_float


def read_field_from_mses(src_file: str, M_inf: float or None = None, gam: float or None = None):
    r"""
    Reads a field dump file from MSES (by default, of the form ``field.*``) and outputs the information to an array.
    The array has shape :math:`9 \times m \times n` (or :math:`12 \times m \times n` if ``M_inf`` and ``gam`` are
    specified), where :math:`m` is the number of streamlines and :math:`n`
    is the number of streamwise cells in a given streamline.

    The 12 entries in the zeroth axis correspond to the flow variables as follows:

    * 0: :math:`x`
    * 1: :math:`y`
    * 2: :math:`\rho/\rho_\infty` (density)
    * 3: :math:`p/p_\infty` (pressure)
    * 4: :math:`u/V_\infty` (:math:`x`-velocity)
    * 5: :math:`v/V_\infty` (:math:`y`-velocity)
    * 6: :math:`|V|/V_\infty` (velocity magnitude)
    * 7: :math:`M` (Mach number)
    * 8: :math:`C_p` (pressure coefficient)
    * 9: :math:`C_{p_t}` (total pressure coefficient)
    * 10: :math:`\Delta C_{p_t}` (change in total pressure coefficient relative to the previous streamwise cell)
    * 11: :math:`\Delta C_p` (change in pressure coefficient relative to the previous streamwise cell)

    Note that for :math:`\Delta C_{p_t}` and :math:`\Delta C_p`, the value of the first streamwise cells (the cells
    along the inlet plane) is defined to be 0. Also note that entries 9, 10, and 11 are only available if ``M_inf``
    and ``gam`` are specified.

    Parameters
    ==========
    src_file: str
        Source file containing the MSES field output

    M_inf: float or None
        Freestream Mach number (must be set if entries 9, 10, or 11 are needed). Default: ``None``.

    gam: float or None
        Specific heat ratio (must be set if entries 9, 10, or 11 are needed). Default: ``None``.

    Returns
    =======
    np.ndarray
        Array of MSES field data, reshaped to :math:`9 \times m \times n`, or :math:`12 \times m \times n` if
        ``M_inf`` and ``gam`` are specified.
    """
    data = np.loadtxt(src_file, skiprows=2)
    n_flow_vars = data.shape[1]

    with open(src_file, "r") as f:
        lines = f.readlines()

    n_streamlines = 0
    for line in lines:
        if line == "\n":
            n_streamlines += 1

    n_streamwise_lines = int(data.shape[0] / n_streamlines)

    field_array = np.array([data[:, i].reshape(n_streamlines, n_streamwise_lines).T for i in range(n_flow_vars)])

    if M_inf is not None and gam is not None:
        Cpt = field_array[3][:, :] * (1 + (gam - 1) / 2 * field_array[7][:, :] ** 2) ** (gam / (gam - 1)) * (2 / gam / M_inf**2)
        Cp = field_array[flow_var_idx["Cp"]][:, :]
        Cpt = np.array([Cpt])
        Cp = np.array([Cp])
        field_array = np.concatenate((field_array, Cpt))
        dCpt = np.zeros(Cpt[0].shape)
        dCp = np.zeros(Cpt[0].shape)
        for idx in range(1, Cpt[0].shape[0]):
            dCpt[idx, :] = Cpt[0][idx, :] - Cpt[0][idx - 1, :]
            dCp[idx, :] = Cp[0][idx, :] - Cp[0][idx - 1, :]
        dCpt = np.array([dCpt])
        dCp = np.array([dCp])
        field_array = np.concatenate((field_array, dCpt))
        field_array = np.concatenate((field_array, dCp))

    return field_array


def read_streamline_grid_from_mses(src_file: str, grid_stats: dict):
    r"""
    Reads the grid of streamlines from an MSES ``grid.*`` file

    Parameters
    ==========
    src_file: str
        File containing the grid output from MPLOT (usually of the form ``grid.*``)

    grid_stats: dict
        Output from the ``read_grid_stats_from_mses()`` function

    Returns
    =======
    typing.Tuple(np.ndarray, np.ndarray)
        Grid x-coordinates and grid y-coordinates split by the stagnation streamlines
    """
    data = np.loadtxt(src_file, skiprows=2)

    with open(src_file, "r") as f:
        lines = f.readlines()

    n_streamlines = 0
    for idx, line in enumerate(lines):
        if line == "\n":
            n_streamlines += 1

    n_streamwise_lines = int(data.shape[0] / n_streamlines)

    x_grid = data[:, 0].reshape(n_streamlines, n_streamwise_lines).T
    y_grid = data[:, 1].reshape(n_streamlines, n_streamwise_lines).T

    streamline_borders = np.sort(grid_stats["Jside2"])
    split_x_grid = np.split(x_grid, streamline_borders, axis=1)
    split_y_grid = np.split(y_grid, streamline_borders, axis=1)

    return split_x_grid, split_y_grid


def convert_blade_file_to_3d_array(src_file: str):
    r"""
    Converts an MSES blade file (by default of the form ``blade.*``) to an array.
    The array has shape :math:`N \times M \times 2`, where :math:`N` is the number of airfoils
    and :math:`M` is the number of discrete airfoil coordinates in a given airfoil.

    Parameters
    ==========
    src_file: str
        Source file containing the MSES blade information (usually ``blade.*``)

    Returns
    =======
    np.ndarray
        Array of the shape :math:`N \times M \times 2`
    """
    blade = np.loadtxt(src_file, skiprows=2)
    airfoil_delimiter_rows = np.argwhere(blade == 999.0)
    array_3d = np.split(blade, np.unique(airfoil_delimiter_rows[:, 0]))
    new_airfoils = [array_3d[0]]
    for airfoil in array_3d[1:]:
        new_airfoil = np.delete(airfoil, 0, axis=0)
        new_airfoils.append(new_airfoil)
    return new_airfoils


def read_polar(airfoil_name: str, base_dir: str) -> typing.Dict[str, typing.List[float]]:
    """
    Reads the ``polar.xxx`` file produced by MPOLAR and converts the data to a dictionary of lists

    Parameters
    ----------
    airfoil_name: str
        Name of the airfoil provided to MSES (also the name of the sub-folder inside ``base_dir`` containing the
        analysis files)

    base_dir: str
        Base directory of the analysis. The file being read should have the form
        ``base_dir/airfoil_name/polar.airfoil_name``

    Returns
    -------
    typing.Dict[str, typing.List[float]]
        Dictionary where each key is a string corresponding to an aerodynamic performance variable and each
        value is a list of the evaluation of that performance variable at every point along the polar.
        For example, the dictionary might look something like ``{"alf": [-1.0, 0.0, 1.0], "Cl": [0.2, 0.3, 0.4], ...}``
    """
    data = np.loadtxt(os.path.join(base_dir, airfoil_name, f"polar.{airfoil_name}"), skiprows=13)
    keys = ["alf", "Cl", "Cd", "Cm", "Cdw", "Cdv", "Cdp", "Ma", "Re", "top_xtr", "bot_xtr"]
    return {k: v.tolist() for (k, v) in zip(keys, data.T)}



import math


class ObjectiveConstraint:
    def __init__(self, func_str: str):
        """
        Objective or Constraint used in shape optimization. Allows for dynamic updates and equation validity checking
        inside the GUI.

        Parameters
        ==========
        func_str: str
            Function string used to define the Objective or Constraint from the ``aero_data`` dictionary output
            from the ``calc_aero_data``.
        """
        self.func_str = func_str
        self.func = None
        self.function_dict = {}
        self.depends_on = {}
        self.value = None
        self.tag_matrix = None
        self.tag_list = None

    def add_or_set_dependencies(self, dependencies: dict):
        """
        Adds performance parameter dependencies from a Python dictionary.

        Parameters
        ==========
        dependencies: dict
            Performance parameters extracted from an airfoil system analysis used to define value of an ``Objective``
            or ``Constraint``.
        """
        for k, v in dependencies.items():
            self.depends_on[k] = v

    def set_func_str(self, func_str: str):
        """
        Simple method that overwrites the ``func_str`` attribute.

        Parameters
        ==========
        func_str: str
            Function string used to define the Objective or Constraint from the ``aero_data`` dictionary output
            from the ``calc_aero_data``.
        """
        self.func_str = func_str

    def remove_func(self):
        """
        Removes the ``Objective`` or ``Constraint`` function from all the relevant locations in the object.
        """
        self.func_str = None
        self.func = None
        self.function_dict = {}

    def update_function(self):
        """
        Updates the function based on the function string and its dependencies.
        """
        if self.func_str is None:
            pass
        else:
            # Convert the function string into a Python function and determine parameters present in string:
            math_function_list = self.parse_update_function_str()

            # Add any math functions detected from the func_str:
            for s in math_function_list:
                if s not in self.function_dict.keys():
                    if s in vars(math).keys():
                        self.function_dict[s] = vars(math)[s]

            # Add the variables the function depends on to the function_dict and detect whether the function should be
            # executed:
            execute = self.add_dependencies()

            # Update the function (not the result) in the function_dict
            if execute:
                try:
                    exec(self.func, self.function_dict)
                    return_val = self.function_dict["f"]()
                    if not isinstance(return_val, float):
                        raise FunctionCompileError(f"Error in function compilation output type. Required type is float,"
                                                   f"found type {type(return_val)}")
                except (SyntaxError, NameError, TypeError):
                    raise FunctionCompileError('Error in function compilation')

    def update_value(self):
        """
        Update the value of the ``Objective`` or ``Constraint`` using the stored function.
        """
        if self.func_str is None:
            pass
        else:
            try:
                self.value = self.function_dict['f']()  # no parameters passed as inputs (inputs all stored and updated
                # inside self.function_dict )
            except (SyntaxError, NameError, TypeError):
                raise FunctionCompileError('Error in function update')

    def update(self, dependencies: dict):
        """
        Updates the function and its value using a set of dependencies.

        Parameters
        ==========
        dependencies: dict
            Performance parameter dependencies from the airfoil system analysis.
        """
        self.add_or_set_dependencies(dependencies)
        self.update_function()
        self.update_value()

    def parse_update_function_str(self):
        """
        Converts the function string to a function executable by Python. The special character "." is used to signal
        a depth increment within the airfoil system hierarchy. The special character "$" is used to define the start of
        a ``Param`` name. For example, the string ``"A0.Base.R_le"`` corresponds to the leading edge radius of the base
        parameter set of airfoil ``"A0"``.
        """
        self.tag_matrix = []
        self.func = 'def f(): return '
        math_functions_to_include = []
        appending = False
        append_new_to_math_function_list = True
        for ch in self.func_str:
            if appending:
                if ch.isalnum() or ch == '_':
                    self.tag_matrix[-1][-1] += ch
                elif ch == '.':
                    self.tag_matrix[-1].append('')
                else:
                    appending = False
            if ch == '$':
                self.tag_matrix.append([''])
                appending = True
            elif ch == '.' and appending:
                self.func += '_'
            else:
                self.func += ch
            if not appending and ch.isalnum():
                if append_new_to_math_function_list:
                    math_functions_to_include.append('')
                math_functions_to_include[-1] += ch
                append_new_to_math_function_list = False
            if not appending and not ch.isalnum():
                append_new_to_math_function_list = True

        def concatenate_strings(lst: list):
            tag = ''
            for idx, s in enumerate(lst):
                tag += s
                if idx < len(lst) - 1:
                    tag += '_'
            return tag

        self.tag_list = [concatenate_strings(tl) for tl in self.tag_matrix]
        for t in self.tag_list:
            self.function_dict[t] = None

        return math_functions_to_include

    def add_dependencies(self):
        """Adds dependencies found in ``depends_on`` to the function dictionary for execution."""

        for idx, t in enumerate(self.tag_list):
            if t in self.depends_on.keys():
                self.function_dict[t] = self.depends_on[t]
            else:
                self.function_dict = {}
                self.remove_func()
                raise FunctionCompileError(f"Could not compile input function string: {self.func_str}")

        return True


class Objective(ObjectiveConstraint):
    def __init__(self, func_str: str):
        """
        Subclass of ``ObjectiveConstraint`` simply specifying the object as an Objective.

        Parameters
        ==========
        func_str: str
            Function string to pass to ``ObjectiveConstraint``'s ``__init__``.
        """
        super().__init__(func_str)


class Constraint(ObjectiveConstraint):
    """
    Subclass of ``ObjectiveConstraint`` simply specifying the object as an Constraint.

    Parameters
    ==========
    func_str: str
        Function string to pass to ``ObjectiveConstraint``'s ``__init__``.
    """
    def __init__(self, func_str: str):
        super().__init__(func_str)


class FunctionCompileError(Exception):
    pass


if __name__ == '__main__':
    obj = Objective('$Cd')
    obj.update({'Cd': 0.02, 'Cl': 0.35, 'Cl_target': 0.08, 'Cl_tol': 0.005})
    constraint = Constraint('abs($Cl - $Cl_target) - $Cl_tol')
    constraint.update({'Cd': 0.02, 'Cl': 0.35, 'Cl_target': 0.3455, 'Cl_tol': 0.005})
    pass



import typing

import numpy as np

from pymead.core.point import PointSequence, Point
from pymead.core.parametric_curve import ParametricCurve, PCurveData
from pymead.utils.get_airfoil import extract_data_from_airfoiltools


class LineSegment(ParametricCurve):

    def __init__(self, point_sequence: PointSequence, name: str or None = None, **kwargs):
        super().__init__(sub_container="lines", **kwargs)
        self._point_sequence = None
        self.set_point_sequence(point_sequence)
        name = "Line-1" if name is None else name
        self.set_name(name)
        self._add_references()

    def _add_references(self):
        for idx, point in enumerate(self.point_sequence().points()):
            # Add the object reference to each point in the curve
            if self not in point.curves:
                point.curves.append(self)

    def point_sequence(self):
        return self._point_sequence

    def set_point_sequence(self, point_sequence: PointSequence):
        if len(point_sequence) != 2:
            raise ValueError("Point sequence must contain exactly two points")
        self._point_sequence = point_sequence

    def reverse_point_sequence(self):
        self.point_sequence().reverse()

    def point_removal_deletes_curve(self):
        return True

    def remove_point(self, idx: int or None = None, point: Point or None = None):
        if isinstance(point, Point):
            idx = self.point_sequence().point_idx_from_ref(point)
        self.point_sequence().remove_point(idx)

        if len(self.point_sequence()) > 1:
            delete_curve = False
        else:
            delete_curve = True

        return delete_curve

    def remove(self):
        if self.canvas_item is not None:
            self.canvas_item.sigRemove.emit(self.canvas_item)

    def update(self):
        p_curve_data = self.evaluate()
        if self.canvas_item is not None:
            self.canvas_item.updateCanvasItem(curve_data=p_curve_data)

    def evaluate(self, t: np.ndarray or None = None, **kwargs):
        if "nt" not in kwargs.keys() and t is None:
            kwargs["nt"] = 2  # Set the default parameter vector for the line to be [0.0, 1.0]
        t = ParametricCurve.generate_t_vec(**kwargs) if t is None else t
        p1 = self.point_sequence().points()[0]
        p2 = self.point_sequence().points()[1]
        x1 = p1.x().value()
        y1 = p1.y().value()
        x2 = p2.x().value()
        y2 = p2.y().value()
        theta = np.arctan2(y2 - y1, x2 - x1)
        r = np.hypot(x2 - x1, y2 - y1)
        x = x1 + t * r * np.cos(theta)
        y = y1 + t * r * np.sin(theta)
        xy = np.column_stack((x, y))
        xpyp = np.repeat(np.array([r * np.cos(theta), r * np.sin(theta)]), t.shape[0])
        xppypp = np.repeat(np.array([0.0, 0.0]), t.shape[0])
        k = np.zeros(t.shape)
        R = np.inf * np.ones(t.shape)
        return PCurveData(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)

    def get_dict_rep(self):
        return {"points": [pt.name() for pt in self.point_sequence().points()]}


class PolyLine(ParametricCurve):

    AirfoilTools = 0
    DatFile = 1

    def __init__(self, source: str, start: int or float = None, end: int or float = None,
                 point_sequence: PointSequence = None, name: str or None = None, num_header_rows: int = 0,
                 delimiter: str or None = None, **kwargs):

        super().__init__(sub_container="polylines", **kwargs)
        self._point_sequence = None
        self.source = source
        self.source_type = self.DatFile if ("/" in source or "\\" in source) else self.AirfoilTools
        self.start = start
        self.end = end
        self.num_header_rows = num_header_rows
        self.delimiter = delimiter
        self.original_coords = self._get_original_coords()
        self.coords = self._get_coord_slice()
        point_sequence = self._extract_point_sequence_from_coords() if point_sequence is None else point_sequence
        self.set_point_sequence(point_sequence)
        name = self._get_default_name() if name is None else name
        self.set_name(name)
        self._add_references()

    def split(self, split: int or Point):
        if split < 3 or split > len(self.coords) - 3:
            raise ValueError("Split value out of bounds for PolyLine")
        polyline1, polyline2 = None, None
        if isinstance(split, int):
            end_1 = self.start + split + 1 if self.start is not None else split + 1
            start_2 = self.start + split if self.start is not None else split
            polyline1 = PolyLine(source=self.source, start=self.start, end=end_1)
            polyline2 = PolyLine(source=self.source, start=start_2, end=self.end)
            polyline2.point_sequence().points()[0] = polyline1.point_sequence().points()[-1]
            polyline2.point_sequence().points()[0].curves.append(polyline2)

            for new_polyline in [polyline1, polyline2]:
                for point_idx, point in enumerate(new_polyline.point_sequence().points()):
                    for original_point in self.point_sequence().points():
                        if not point.is_coincident(original_point):
                            continue
                        new_polyline.point_sequence().points()[point_idx] = original_point
                        if self in original_point.curves:
                            original_point.curves.remove(self)
                        if new_polyline not in original_point.curves:
                            original_point.curves.append(new_polyline)
                        break

        return [polyline1, polyline2]

    def add_polyline_airfoil(self):
        le = self._add_le()
        te = self._add_te()
        blunt_trailing_edge = self._add_trailing_edge_lines(te)
        self._add_airfoil(le, te, blunt_trailing_edge=blunt_trailing_edge)

    def _add_le(self):
        coords_dist_from_origin = np.hypot(self.coords[:, 0], self.coords[:, 1])
        le_row = np.argmin(coords_dist_from_origin)
        le = self.geo_col.add_point(self.coords[le_row, 0], self.coords[le_row, 1])
        le.curves.append(self)
        return le

    def _add_te(self):
        if len(self.point_sequence().points()) == 1:
            return self.point_sequence().points()[0]
        if self.coords[0, 1] >= 0.0 >= self.coords[-1, 1]:
            te = self.geo_col.add_point(1.0, 0.0)
        else:
            te = self.geo_col.add_point(0.5 * (self.coords[0, 0] + self.coords[-1, 0]),
                                        0.5 * (self.coords[0, 1] + self.coords[-1, 1]))
        return te

    def _add_trailing_edge_lines(self, te: Point):
        if len(self.point_sequence().points()) == 1:
            return False
        self.geo_col.add_line(PointSequence(points=[te, self.point_sequence().points()[0]]))
        self.geo_col.add_line(PointSequence(points=[te, self.point_sequence().points()[-1]]))
        return True

    def _add_airfoil(self, le: Point, te: Point, blunt_trailing_edge: bool):
        if blunt_trailing_edge:
            self.geo_col.add_airfoil(le, te,
                                     upper_surf_end=self.point_sequence().points()[0],
                                     lower_surf_end=self.point_sequence().points()[-1])
        else:
            self.geo_col.add_airfoil(le, te, upper_surf_end=te, lower_surf_end=te)

    def _get_default_name(self):
        if self.source_type == self.AirfoilTools:
            return f"{self.source}-1"
        return "PolyLine-1"

    def _get_original_coords(self):
        if self.source_type == self.AirfoilTools:
            return self._load_coords_from_airfoil_tools()
        elif self.source_type == self.DatFile:
            try:
                return self._load_coords_from_dat_file(self.num_header_rows, self.delimiter)
            except:
                self.num_header_rows = 1
                return self._load_coords_from_dat_file(self.num_header_rows, self.delimiter)
        else:
            raise ValueError("Invalid polyline source type")

    def _get_coord_slice(self):
        if self.start is None and isinstance(self.end, int):
            return self.original_coords[:self.end, :]
        elif self.end is None and isinstance(self.start, int):
            return self.original_coords[self.start:, :]
        elif isinstance(self.start, int) and isinstance(self.end, int):
            return self.original_coords[self.start:self.end, :]
        else:
            return self.original_coords

    def _load_coords_from_dat_file(self, num_header_rows: int = 0, delimiter: str or None = None):
        return np.loadtxt(self.source, skiprows=num_header_rows, delimiter=delimiter)

    def _load_coords_from_airfoil_tools(self):
        return extract_data_from_airfoiltools(self.source)

    def _extract_point_sequence_from_coords(self):
        if Point(self.coords[0, 0], self.coords[0, 1]).is_coincident(Point(self.coords[-1, 0], self.coords[-1, 1])):
            return PointSequence(points=[Point(self.coords[0, 0], self.coords[0, 1])])
        return PointSequence(points=[Point(self.coords[row, 0], self.coords[row, 1]) for row in [0, -1]])

    def _add_references(self):
        for idx, point in enumerate(self.point_sequence().points()):
            # Add the object reference to each point in the curve
            if self not in point.curves:
                point.curves.append(self)

    def point_sequence(self):
        return self._point_sequence

    def set_point_sequence(self, point_sequence: PointSequence):
        self._point_sequence = point_sequence

    def reverse_point_sequence(self):
        self.point_sequence().reverse()

    def point_removal_deletes_curve(self):
        return True

    def remove_point(self, idx: int or None = None, point: Point or None = None):
        if isinstance(point, Point):
            idx = self.point_sequence().point_idx_from_ref(point)
        self.point_sequence().remove_point(idx)

        if len(self.point_sequence()) > 1:
            delete_curve = False
        else:
            delete_curve = True

        return delete_curve

    def remove(self):
        if self.canvas_item is not None:
            self.canvas_item.sigRemove.emit(self.canvas_item)

    def update(self):
        p_curve_data = self.evaluate()
        if self.canvas_item is not None:
            self.canvas_item.updateCanvasItem(curve_data=p_curve_data)

    def evaluate(self, t: np.ndarray or None = None, **kwargs):
        xy = self.coords
        t = np.linspace(0.0, 1.0, xy.shape[0])
        xp = np.gradient(xy[:, 0], t)
        yp = np.gradient(xy[:, 1], t)
        xpyp = np.column_stack((xp, yp))
        xpp = np.gradient(xp, t)
        ypp = np.gradient(yp, t)
        xppypp = np.column_stack((xpp, ypp))
        with np.errstate(divide="ignore"):
            k = np.true_divide(xpyp[:, 0] * xppypp[:, 1] - xpyp[:, 1] * xppypp[:, 0],
                               np.hypot(xpyp[:, 0], xpyp[:, 1])**1.5)
            R = np.true_divide(1, k)
        return PCurveData(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)

    def get_dict_rep(self):
        return {"points": [pt.name() for pt in self.point_sequence().points()], "source": self.source,
                "start": self.start, "end": self.end}



from PyQt5.QtWidgets import QWidget, QHBoxLayout, QProgressBar, QLabel, QComboBox

from pymead.version import __version__


class PymeadProgressBar(QProgressBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        # self.setStyleSheet(
        #     '''
        #     QProgressBar {border: 2px solid; border-color: #8E9091;}
        #     QProgressBar::chunk {background-color: #6495ED; width: 10px; margin: 0.5px;}
        #     '''
        # )
        self.setMaximum(100)
        self.setValue(0)
        self.hide()


class PymeadInfoLabel(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setText(f"pymead {__version__}")


class PermanentWidget(QWidget):
    def __init__(self, geo_col, parent=None):
        super().__init__(parent)
        self.geo_col = geo_col
        self.lay = QHBoxLayout(self)
        self.inviscid_cl_label = QLabel("Inviscid CL", self)
        self.inviscid_cl_combo = QComboBox(self)
        self.inviscid_cl_combo.setMinimumWidth(80)
        self.progress_bar = PymeadProgressBar(self)
        self.info_label = PymeadInfoLabel(self)
        self.lay.addWidget(self.inviscid_cl_label)
        self.lay.addWidget(self.inviscid_cl_combo)
        self.lay.addWidget(self.progress_bar)
        self.lay.addWidget(self.info_label)
        self.setLayout(self.lay)

    def updateAirfoils(self):
        previous_item = self.inviscid_cl_combo.currentText()
        new_item_list = [""] + [a for a in self.geo_col.container()["airfoils"]]

        # Clear the combo box and add the new items
        self.inviscid_cl_combo.clear()
        self.inviscid_cl_combo.addItems(new_item_list)

        # If the previous airfoil is still present, set it as the current selected item. This will
        # prevent the current airfoil being analyzed from changing when airfoils are added
        if previous_item in new_item_list:
            self.inviscid_cl_combo.setCurrentText(previous_item)



import matplotlib.pyplot as plt
from matplotlib import colors as mpl_colors
import numpy as np
import os
from pymead.analysis.read_aero_data import read_grid_stats_from_mses, read_field_from_mses, \
    read_streamline_grid_from_mses, read_Mach_from_mses_file
from pymead.analysis.read_aero_data import flow_var_idx


flow_var_label = {'M': 'Mach Number',
                  'Cp': 'Pressure Coefficient',
                  'p': 'Static Pressure (p / p<sub>\u221e</sub>)',
                  'rho': 'Density (\u03c1/\u03c1<sub>\u221e</sub>)',
                  'u': 'Velocity-x (u/V<sub>\u221e</sub>)',
                  'v': 'Velocity-y (v/V<sub>\u221e</sub>)',
                  'V': 'Velocity-mag (V/V<sub>\u221e</sub>)',
                  'q': 'Speed of Sound (q/V<sub>\u221e</sub>)',
                  "Cpt": "Total Pressure / Pinf",
                  "dCpt": "Delta Total Pressure"}


def generate_field_matplotlib(axs: plt.Axes or None, analysis_subdir: str, var: str, cmap_field: mpl_colors.Colormap or str ,
                              cmap_airfoil: mpl_colors.Colormap or str, shading: str = 'gouraud', vmin: float = None,
                              vmax: float = None, arrow_start_J_idx: int = 20, arrow_spacing: int = 60):
    pcolormesh_handles = {'field': None, 'airfoil': []}
    field_file = os.path.join(analysis_subdir, f'field.{os.path.split(analysis_subdir)[-1]}')
    grid_stats_file = os.path.join(analysis_subdir, 'mplot_grid_stats.log')
    grid_file = os.path.join(analysis_subdir, f'grid.{os.path.split(analysis_subdir)[-1]}')
    mses_file = os.path.join(analysis_subdir, f"mses.{os.path.split(analysis_subdir)[-1]}")
    # bl_file = os.path.join(analysis_subdir, f'bl.{os.path.split(analysis_subdir)[-1]}')
    if not os.path.exists(field_file):
        raise OSError(f"Field file {field_file} not found")
    if not os.path.exists(grid_stats_file):
        raise OSError(f"Grid statistics log {grid_stats_file} not found")
    if not os.path.exists(grid_file):
        raise OSError(f"Grid file {grid_file} not found")

    # bl_data = read_bl_data_from_mses(bl_file)
    M_inf = read_Mach_from_mses_file(mses_file)
    gam = 1.4
    field = read_field_from_mses(field_file, M_inf=M_inf, gam=gam)
    grid_stats = read_grid_stats_from_mses(grid_stats_file)
    x_grid, y_grid = read_streamline_grid_from_mses(grid_file, grid_stats)

    flow_var = field[flow_var_idx[var]]

    if vmin is None:
        vmin = np.min(flow_var)
    if vmax is None:
        vmax = np.max(flow_var)

    # Stagnation streamline arrows
    max_arrow_idx = x_grid[0].shape[0] - 2
    if arrow_start_J_idx > max_arrow_idx:
        raise ValueError(f"Starting arrow index too large (greater than max J index of {max_arrow_idx})")
    arrow_idxs = [arrow_start_J_idx]
    while True:
        if arrow_idxs[-1] + arrow_spacing > max_arrow_idx:
            break
        else:
            arrow_idxs.append(arrow_idxs[-1] + arrow_spacing)

    start_idx, end_idx = 0, x_grid[0].shape[1] - 1
    for flow_section_idx in range(grid_stats["numel"] + 1):

        flow_var_section = flow_var[:, start_idx:end_idx]

        args = (x_grid[flow_section_idx], y_grid[flow_section_idx], flow_var_section)
        kwargs = dict(cmap=cmap_field, vmin=vmin, vmax=vmax)

        if axs is None:
            pcolormesh_handles[f'field_{flow_section_idx}'] = plt.pcolormesh(*args, **kwargs)
        else:
            pcolormesh_handles[f'field_{flow_section_idx}'] = axs.pcolormesh(*args, **kwargs)

        if flow_section_idx == 0:
            x_stags = [x_grid[flow_section_idx][:, -1]]
            y_stags = [y_grid[flow_section_idx][:, -1]]
        elif flow_section_idx == grid_stats["numel"]:
            x_stags = [x_grid[flow_section_idx][:, 0]]
            y_stags = [y_grid[flow_section_idx][:, 0]]
        else:
            x_stags = [x_grid[flow_section_idx][:, 0], x_grid[flow_section_idx][:, -1]]
            y_stags = [y_grid[flow_section_idx][:, 0], y_grid[flow_section_idx][:, -1]]

        for x_stag, y_stag in zip(x_stags, y_stags):
            axs.plot(x_stag, y_stag, color="blue")
            for arrow_idx in arrow_idxs:
                axs.arrow(x_stag[arrow_idx], y_stag[arrow_idx],
                          x_stag[arrow_idx + 1] - x_stag[arrow_idx], y_stag[arrow_idx + 1] - y_stag[arrow_idx],
                          color="blue", width=0.0001, head_width=0.01, zorder=1000)

        if flow_section_idx < grid_stats["numel"]:
            start_idx = end_idx
            end_idx += x_grid[flow_section_idx + 1].shape[1] - 1

    # for bl in bl_data:
    #     axs.plot(bl["x"], bl["y"], color="white", marker="x", mfc="none", mec="white")

    if axs is None:
        return pcolormesh_handles
    else:
        return axs, pcolormesh_handles



from abc import abstractmethod

import pyqtgraph as pg
from PyQt5.QtCore import Qt, pyqtSignal, QObject
from PyQt5.QtGui import QFont, QPen, QColor

from pymead.core import UNITS
from pymead.core.constraints import *
from pymead.utils.misc import get_setting


class ConstraintItem(QObject):

    sigItemHovered = pyqtSignal()
    sigItemLeaveHovered = pyqtSignal()

    def __init__(self, constraint: GeoCon, canvas_items: list, theme: dict):
        super().__init__()
        self.constraint = constraint
        self.constraint.canvas_item = self
        self.canvas_items = canvas_items
        self.setStyle(theme=theme, mode="default")

    def addItems(self, canvas):
        for item in self.canvas_items:
            canvas.addItem(item)
            # TODO: connect hover leave/enter signals here

    def hide(self):
        for canvas_item in self.canvas_items:
            canvas_item.hide()

    def show(self):
        for canvas_item in self.canvas_items:
            canvas_item.show()

    @abstractmethod
    def update(self):
        pass

    def setStyle(self, theme: dict = None, mode: str = "default"):
        if mode == "default" and theme is None:
            raise ValueError("Must specify theme for default constraint style")
        for item in self.canvas_items:
            if isinstance(item, pg.ArrowItem):
                if mode == "default":
                    item.setStyle(brush=theme["main-color"])
                else:
                    item.setStyle(brush=get_setting(f"curve_{mode}_pen_color"))
            elif isinstance(item, pg.TextItem):
                if mode == "default":
                    item.setColor(theme["main-color"])
                else:
                    item.setColor(get_setting(f"curve_{mode}_pen_color"))
            elif isinstance(item, pg.PlotDataItem) or isinstance(item, pg.PlotCurveItem):
                pen = item.opts["pen"]
                if isinstance(pen, QPen):
                    if mode == "default":
                        pen.setColor(QColor(theme["main-color"]))
                    else:
                        pen.setColor(QColor(get_setting(f"curve_{mode}_pen_color")))
                item.setPen(pen)


class DistanceConstraintItem(ConstraintItem):
    def __init__(self, constraint: DistanceConstraint, theme: dict):
        self.arrow_style = {"headLen": 10}
        self.text_style = {"anchor": (0.5, 0.5)}
        canvas_items = [
            pg.ArrowItem(**self.arrow_style),
            pg.ArrowItem(**self.arrow_style),
            pg.TextItem(**self.text_style),
            pg.PlotCurveItem(),
            pg.PlotCurveItem(),
            pg.PlotCurveItem(),
        ]
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        self.canvas_items[2].setFont(QFont("DejaVu Sans Mono", 10))
        for item in canvas_items:
            if isinstance(item, pg.ArrowItem):
                item.setZValue(-8)
            else:
                item.setZValue(-10)
        self.update()

    def update(self):
        angle = self.constraint.p1.measure_angle(self.constraint.p2)
        dist = self.constraint.p1.measure_distance(self.constraint.p2)

        handle_angle = angle + np.pi / 2
        arrow_offset = 0.04 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())
        p1_arrow = (self.constraint.p1.x().value() + arrow_offset * np.cos(handle_angle),
                    self.constraint.p1.y().value() + arrow_offset * np.sin(handle_angle))
        p2_arrow = (self.constraint.p2.x().value() + arrow_offset * np.cos(handle_angle),
                    self.constraint.p2.y().value() + arrow_offset * np.sin(handle_angle))

        text_offset = 0.06 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())
        text_angle = np.rad2deg((angle + np.pi / 2) % np.pi - np.pi / 2)
        text_pos = (self.constraint.p1.x().value() + text_offset * np.cos(handle_angle) + 0.5 * dist * np.cos(angle),
                    self.constraint.p1.y().value() + text_offset * np.sin(handle_angle) + 0.5 * dist * np.sin(angle))

        line_x = [self.constraint.p1.x().value() + arrow_offset * np.cos(handle_angle),
                  self.constraint.p1.x().value() + arrow_offset * np.cos(handle_angle) + dist * np.cos(angle)]
        line_y = [self.constraint.p1.y().value() + arrow_offset * np.sin(handle_angle),
                  self.constraint.p1.y().value() + arrow_offset * np.sin(handle_angle) + dist * np.sin(angle)]

        handle_offset = 0.05 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())
        handle1_x = [self.constraint.p1.x().value(),
                     self.constraint.p1.x().value() + handle_offset * np.cos(handle_angle)]
        handle1_y = [self.constraint.p1.y().value(),
                     self.constraint.p1.y().value() + handle_offset * np.sin(handle_angle)]

        handle2_x = [self.constraint.p2.x().value(),
                     self.constraint.p2.x().value() + handle_offset * np.cos(handle_angle)]
        handle2_y = [self.constraint.p2.y().value(),
                     self.constraint.p2.y().value() + handle_offset * np.sin(handle_angle)]

        self.canvas_items[0].setPos(*p1_arrow)
        self.canvas_items[0].setStyle(angle=-np.rad2deg(angle))
        self.canvas_items[1].setPos(*p2_arrow)
        self.canvas_items[1].setStyle(angle=-np.rad2deg(angle) + 180)
        self.canvas_items[2].setPos(*text_pos)
        self.canvas_items[2].setText(f"{self.constraint.param().value():.4f}")
        self.canvas_items[2].setAngle(text_angle)
        self.canvas_items[3].setData(x=line_x, y=line_y)
        self.canvas_items[4].setData(x=handle1_x, y=handle1_y)
        self.canvas_items[5].setData(x=handle2_x, y=handle2_y)


class SymmetryConstraintItem(ConstraintItem):
    def __init__(self, constraint: SymmetryConstraint, theme: dict):
        canvas_items = [
            pg.TextItem("\u24c2"),
            pg.TextItem("\u24c2")
        ]
        canvas_items[0].setFont(QFont("DejaVu Sans", 10))
        canvas_items[1].setFont(QFont("DejaVu Sans", 10))
        for item in canvas_items:
            item.setZValue(-10)
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        self.update()

    def update(self):
        dist = 0.01 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())
        self.canvas_items[0].setPos(self.constraint.p3.x().value() + dist, self.constraint.p3.y().value() - dist)
        self.canvas_items[1].setPos(self.constraint.p4.x().value() + dist, self.constraint.p4.y().value() - dist)


class ROCurvatureConstraintItem(ConstraintItem):
    def __init__(self, constraint: ROCurvatureConstraint, theme: dict):
        canvas_items = [
            pg.TextItem("\u24c7"),
        ]
        canvas_items[0].setFont(QFont("DejaVu Sans", 10))
        for item in canvas_items:
            item.setZValue(-10)
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        self.update()

    def update(self):
        dist = 0.01 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())
        self.canvas_items[0].setPos(self.constraint.curve_joint.x().value() - dist,
                                    self.constraint.curve_joint.y().value() - dist)


class AntiParallel3ConstraintItem(ConstraintItem):
    def __init__(self, constraint: AntiParallel3Constraint, theme: dict):
        self.text_style = dict(anchor=(0.5, 0.5))
        pen = pg.mkPen(width=1, style=Qt.DashLine)
        canvas_items = [
            pg.TextItem("\u2225", **self.text_style),
            pg.TextItem("\u2225", **self.text_style),
            pg.PlotDataItem(pen=pen),
            pg.PlotDataItem(pen=pen)
        ]
        canvas_items[0].setFont(QFont("DejaVu Sans", 10))
        canvas_items[1].setFont(QFont("DejaVu Sans", 10))
        for item in canvas_items:
            item.setZValue(-10)
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        self.update()

    def update(self):
        midpoint1 = [np.mean([self.constraint.p1.x().value(), self.constraint.p2.x().value()]),
                     np.mean([self.constraint.p1.y().value(), self.constraint.p2.y().value()])]
        midpoint2 = [np.mean([self.constraint.p3.x().value(), self.constraint.p2.x().value()]),
                     np.mean([self.constraint.p3.y().value(), self.constraint.p2.y().value()])]
        line1_x = [self.constraint.p2.x().value(), self.constraint.p1.x().value()]
        line1_y = [self.constraint.p2.y().value(), self.constraint.p1.y().value()]

        line2_x = [self.constraint.p2.x().value(), self.constraint.p3.x().value()]
        line2_y = [self.constraint.p2.y().value(), self.constraint.p3.y().value()]
        self.canvas_items[0].setPos(*midpoint1)
        self.canvas_items[1].setPos(*midpoint2)
        self.canvas_items[2].setData(x=line1_x, y=line1_y)
        self.canvas_items[3].setData(x=line2_x, y=line2_y)


class RelAngle3ConstraintItem(ConstraintItem):
    def __init__(self, constraint: RelAngle3Constraint, theme: dict):
        pen = pg.mkPen(width=1, style=Qt.DashLine)
        canvas_items = [
            pg.PlotDataItem(),
            pg.PlotDataItem(pen=pen),
            pg.PlotDataItem(pen=pen),
            pg.TextItem(anchor=(0, 0.5)),
        ]
        canvas_items[3].setFont(QFont("DejaVu Sans Mono", 10))
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        for item in canvas_items:
            item.setZValue(-10)
        self.update()

    def update(self):
        dist1 = self.constraint.p2.measure_distance(self.constraint.p1)
        dist2 = self.constraint.p2.measure_distance(self.constraint.p3)
        angle1 = self.constraint.p2.measure_angle(self.constraint.p1)
        angle2 = angle1 - self.constraint.param().rad()
        mean_angle = np.mean([angle1, angle2])
        text_distance = 0.15 * np.mean([dist1, dist2])
        text_x = self.constraint.p2.x().value() + text_distance * np.cos(mean_angle)
        text_y = self.constraint.p2.y().value() + text_distance * np.sin(mean_angle)

        theta = np.linspace(angle1, angle2, 30)
        x = self.constraint.p2.x().value() + np.mean([dist1, dist2]) * 0.1 * np.cos(theta)
        y = self.constraint.p2.y().value() + np.mean([dist1, dist2]) * 0.1 * np.sin(theta)

        line1_x = [self.constraint.p2.x().value(), self.constraint.p1.x().value()]
        line1_y = [self.constraint.p2.y().value(), self.constraint.p1.y().value()]

        line2_x = [self.constraint.p2.x().value(), self.constraint.p3.x().value()]
        line2_y = [self.constraint.p2.y().value(), self.constraint.p3.y().value()]

        self.canvas_items[0].setData(x=x, y=y)
        self.canvas_items[1].setData(x=line1_x, y=line1_y)
        self.canvas_items[2].setData(x=line2_x, y=line2_y)
        self.canvas_items[3].setPos(text_x, text_y)
        text = f"{self.constraint.param().value():.2f}"
        if self.constraint.param().unit() == "deg":
            text += "\u00b0"
        self.canvas_items[3].setText(text)


class Perp3ConstraintItem(ConstraintItem):
    def __init__(self, constraint: Perp3Constraint, theme: dict):
        pen = pg.mkPen(width=1, style=Qt.DashLine)
        canvas_items = [
            pg.PlotDataItem(),
            pg.PlotDataItem(),
            pg.PlotDataItem(pen=pen),
            pg.PlotDataItem(pen=pen)
        ]
        super().__init__(constraint=constraint, canvas_items=canvas_items, theme=theme)
        for item in canvas_items:
            item.setZValue(-10)
        self.update()

    def update(self):
        angle1 = self.constraint.p2.measure_angle(self.constraint.p1)
        angle2 = angle1 - np.pi / 2

        square_side = 0.05 * UNITS.convert_length_from_base(1.0, UNITS.current_length_unit())

        x1 = [self.constraint.p2.x().value() + square_side * np.cos(angle1),
              self.constraint.p2.x().value() + square_side * np.cos(angle1) + square_side * np.cos(angle2)]
        y1 = [self.constraint.p2.y().value() + square_side * np.sin(angle1),
              self.constraint.p2.y().value() + square_side * np.sin(angle1) + square_side * np.sin(angle2)]
        x2 = [self.constraint.p2.x().value() + square_side * np.cos(angle2),
              self.constraint.p2.x().value() + square_side * np.cos(angle2) + square_side * np.cos(angle1)]
        y2 = [self.constraint.p2.y().value() + square_side * np.sin(angle2),
              self.constraint.p2.y().value() + square_side * np.sin(angle2) + square_side * np.sin(angle1)]

        line1_x = [self.constraint.p2.x().value(), self.constraint.p1.x().value()]
        line1_y = [self.constraint.p2.y().value(), self.constraint.p1.y().value()]

        line2_x = [self.constraint.p2.x().value(), self.constraint.p3.x().value()]
        line2_y = [self.constraint.p2.y().value(), self.constraint.p3.y().value()]

        self.canvas_items[0].setData(x=x1, y=y1)
        self.canvas_items[1].setData(x=x2, y=y2)
        self.canvas_items[2].setData(x=line1_x, y=line1_y)
        self.canvas_items[3].setData(x=line2_x, y=line2_y)



import os

import pyqtgraph as pg
from PyQt5.QtCore import Qt, pyqtSignal
import numpy as np

from pymead import q_settings, GUI_SETTINGS_DIR
from pymead.utils.read_write_files import load_data

q_settings_descriptions = load_data(os.path.join(GUI_SETTINGS_DIR, "q_settings_descriptions.json"))


class DraggablePoint(pg.GraphItem):
    sigPointClicked = pyqtSignal(object, object, object, object)
    sigPointHovered = pyqtSignal(object, object, object, object)
    sigPointLeaveHovered = pyqtSignal(object, object, object, object)
    sigPointStartedMoving = pyqtSignal(object)
    sigPointMoving = pyqtSignal(object)
    sigPointFinishedMoving = pyqtSignal(object)

    def __init__(self):
        self.point = None
        self.dragPoint = None
        self.dragOffset = None
        self.curveOwners = []
        self.textItems = []
        super().__init__()
        self.scatter.sigClicked.connect(self.clicked)
        self.scatter.sigHovered.connect(self.hovered)
        self.hoverable = True

    def setData(self, **kwds):
        self.text = kwds.pop('text', [])
        self.data = kwds
        if 'pos' in self.data:
            npts = self.data['pos'].shape[0]
            self.data['data'] = np.empty(npts, dtype=[('index', int)])
            self.data['data']['index'] = np.arange(npts)
        self.setTexts(self.text)
        self.updateGraph()

    def updateCanvasItem(self, x: float, y: float):
        self.data["pos"][0] = np.array([x, y])
        self.updateGraph()

    def updateGraph(self):
        pg.GraphItem.setData(self, **self.data)
        for i, item in enumerate(self.textItems):
            item.setPos(*self.data['pos'][i])
        self.sigPointMoving.emit(self)

    def setTexts(self, text):
        for i in self.textItems:
            i.scene().removeItem(i)
        self.textItems = []
        for t in text:
            item = pg.TextItem(t)
            self.textItems.append(item)
            item.setParentItem(self)

    def mouseDragEvent(self, ev):
        if ev.button() != Qt.MouseButton.LeftButton:
            ev.ignore()
            return

        if ev.isStart():
            self.sigPointStartedMoving.emit(self)
            # While dragging, disable hovering for curves
            self.hoverable = False
            for curve in self.curveOwners:
                curve.hoverable = False

            # We are already one step into the drag.
            # Find the point(s) at the mouse cursor when the button was first
            # pressed:
            pos = ev.buttonDownPos()
            pts = self.scatter.pointsAt(pos)
            if len(pts) == 0:
                ev.ignore()
                return
            self.dragPoint = pts[0]
            ind = pts[0].data()[0]
            self.dragOffset = self.data['pos'][ind] - pos
        elif ev.isFinish():
            self.sigPointFinishedMoving.emit(self)
            # Re-enable hovering for curves
            self.hoverable = True
            for curve in self.curveOwners:
                curve.hoverable = True

            self.dragPoint = None
            return
        else:
            if self.dragPoint is None:
                ev.ignore()
                return

        data = ev.pos() + self.dragOffset
        x = data.x()
        y = data.y()

        # Make a request to the API to move the point. The GUI representation of the point will move if successful
        self.point.request_move(x, y)

        ev.accept()

    # @staticmethod
    # def hover_tip(x: float, y: float, data):
    #     """
    #     Shows data about a point when it is hovered
    #
    #     Parameters
    #     ==========
    #     x: float
    #         x-location of the control point
    #
    #     y: float
    #         y-location of the control point
    #
    #     data
    #         Signaled by the hover
    #     """
    #     return f"x: {x:.8f}\ny: {y:.8f}\n{data}"

    def clicked(self, item, spot, ev):
        self.sigPointClicked.emit(item, spot, ev, self)

    def hovered(self, item, spot, ev):
        if self.hoverable:
            if ev.exit:
                self.sigPointLeaveHovered.emit(item, spot, ev, self)
            else:
                self.sigPointHovered.emit(item, spot, ev, self)

    def setScatterStyle(self, mode: str = "default"):
        implemented_style_modes = ["default", "hovered", "selected"]
        if mode not in implemented_style_modes:
            raise NotImplementedError(f"Selected mode ({mode}) not implemented. "
                                      f"Currently implemented modes: {implemented_style_modes}.")
        self.scatter.setPen(pg.mkPen(color=q_settings.value(f"scatter_{mode}_pen_color",
                                                            q_settings_descriptions[f"scatter_{mode}_pen_color"][1])))
        self.scatter.setBrush(
            pg.mkBrush(color=q_settings.value(f"scatter_{mode}_brush_color",
                                              q_settings_descriptions[f"scatter_{mode}_brush_color"][1]))
        )
        self.scatter.setSize(q_settings.value(f"scatter_{mode}_size",
                                              q_settings_descriptions[f"scatter_{mode}_size"][1]))
        self.scatter.setSymbol(q_settings.value(f"scatter_{mode}_symbol",
                                                q_settings_descriptions[f"scatter_{mode}_symbol"][1]))



import numpy as np


class Transformation2D:
    def __init__(self, tx: list = None, ty: list = None, r: list = None, sx: list = None,
                 sy: list = None, rotation_units: str = 'rad', order='r,s,t'):
        """Allows for arbitrary 2D transformations on a set of coordinates of size 2 x N"""

        self.tx = tx if tx is not None else [0.0]
        self.ty = ty if ty is not None else [0.0]
        self.r = r if r is not None else [0.0]
        self.sx = sx if sx is not None else [1.0]
        self.sy = sy if sy is not None else [1.0]

        if rotation_units == 'rad':
            self.r_rad = self.r
            self.r_deg = np.rad2deg(np.array(self.r)).tolist()
        elif rotation_units == 'deg':
            self.r_deg = self.r
            self.r_rad = np.deg2rad(np.array(self.r)).tolist()

        self.order = order

        self.r_mat = None
        self.s_mat = None
        self.t_mat = None

        self.M = np.eye(3)  # 3 x 3 identity matrix

        self.generate_rotation_matrix()
        self.generate_scale_matrix()
        self.generate_translation_matrix()
        self.generate_transformation_matrix()

    def generate_rotation_matrix(self):
        self.r_mat = np.array([np.array([[np.cos(r), -np.sin(r), 0],
                                         [np.sin(r), np.cos(r), 0],
                                         [0, 0, 1]]) for r in self.r_rad])

    def generate_scale_matrix(self):
        self.s_mat = np.array([np.array([[self.sx[idx], 0, 0],
                                         [0, self.sy[idx], 0],
                                         [0, 0, 1]]) for idx in range(len(self.sx))])

    def generate_translation_matrix(self):
        self.t_mat = np.array([np.array([[1, 0, self.tx[idx]],
                                         [0, 1, self.ty[idx]],
                                         [0, 0, 1]]) for idx in range(len(self.tx))])

    def generate_transformation_matrix(self):
        r_count = 0
        s_count = 0
        t_count = 0
        for idx, operation in enumerate(self.order.split(',')):
            if operation == 'r':
                self.M = self.M @ self.r_mat[r_count].T
                r_count += 1
            elif operation == 's':
                self.M = self.M @ self.s_mat[s_count].T
                s_count += 1
            elif operation == 't':
                self.M = self.M @ self.t_mat[t_count].T
                t_count += 1
            else:
                raise TransformationError(f'Invalid value \'{operation}\' found in 2-D transformation '
                                          f'(must be \'r\', \'s\', or \'t\'')

    def transform(self, coordinates: np.ndarray):
        """Computes the transformation of the coordinates.

        Parameters
        ==========
        coordinates: np.ndarray
          Size N x 2, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return (np.column_stack((coordinates, np.ones(len(coordinates)))) @ self.M)[:, :2]  # x' = x * M


class Transformation3D:
    def __init__(self, tx: list = None, ty: list = None, tz: list = None, rx: list = None, ry: list = None,
                 rz: list = None, sx: list = None, sy: list = None, sz: list = None,
                 rotation_units: str = 'rad', order='rx,ry,rz,s,t'):
        """Allows for arbitrary 3D transformations on a set of coordinates of size 3 x N"""

        self.tx = tx if tx is not None else [0.0]
        self.ty = ty if ty is not None else [0.0]
        self.tz = tz if tz is not None else [0.0]
        self.rx = rx if rx is not None else [0.0]
        self.ry = ry if ry is not None else [0.0]
        self.rz = rz if rz is not None else [0.0]
        self.sx = sx if sx is not None else [1.0]
        self.sy = sy if sy is not None else [1.0]
        self.sz = sz if sz is not None else [1.0]

        if rotation_units == 'rad':
            self.rx_rad = self.rx
            self.rx_deg = np.rad2deg(np.array(self.rx)).tolist()
            self.ry_rad = self.ry
            self.ry_deg = np.rad2deg(np.array(self.ry)).tolist()
            self.rz_rad = self.rz
            self.rz_deg = np.rad2deg(np.array(self.rz)).tolist()
        elif rotation_units == 'deg':
            self.rx_deg = self.rx
            self.rx_rad = np.deg2rad(np.array(self.rx)).tolist()
            self.ry_deg = self.ry
            self.ry_rad = np.deg2rad(np.array(self.ry)).tolist()
            self.rz_deg = self.rz
            self.rz_rad = np.deg2rad(np.array(self.rz)).tolist()

        self.order = order.split(',')

        self.rx_mat = None
        self.ry_mat = None
        self.rz_mat = None
        self.s_mat = None
        self.t_mat = None

        self.M = np.eye(4)  # 4 x 4 identity matrix

        self.generate_rotation_matrix_x()
        self.generate_rotation_matrix_y()
        self.generate_rotation_matrix_z()
        self.generate_scale_matrix()
        self.generate_translation_matrix()
        self.generate_transformation_matrix()

    def generate_rotation_matrix_x(self):
        self.rx_mat = np.array([np.array([[1, 0, 0, 0],
                                          [0, np.cos(r), np.sin(r), 0],
                                          [0, -np.sin(r), np.cos(r), 0],
                                          [0, 0, 0, 1]]
                                         ) for r in self.rx_rad])

    def generate_rotation_matrix_y(self):
        self.ry_mat = np.array([np.array([[np.cos(r), 0, -np.sin(r), 0],
                                          [0, 1, 0, 0],
                                          [np.sin(r), 0, np.cos(r), 0],
                                          [0, 0, 0, 1]]
                                         ) for r in self.ry_rad])

    def generate_rotation_matrix_z(self):
        self.rz_mat = np.array([np.array([[np.cos(r), -np.sin(r), 0, 0],
                                          [np.sin(r), np.cos(r), 0, 0],
                                          [0, 0, 1, 0],
                                          [0, 0, 0, 1]]
                                         ) for r in self.rz_rad])

    def generate_scale_matrix(self):
        self.s_mat = np.array([np.array([[self.sx[idx], 0, 0, 0],
                                         [0, self.sy[idx], 0, 0],
                                         [0, 0, self.sz[idx], 0],
                                         [0, 0, 0, 1]]) for idx in range(len(self.sx))])

    def generate_translation_matrix(self):
        self.t_mat = np.array([np.array([[1, 0, 0, self.tx[idx]],
                                         [0, 1, 0, self.ty[idx]],
                                         [0, 0, 1, self.tz[idx]],
                                         [0, 0, 0, 1]]) for idx in range(len(self.tx))])

    def generate_transformation_matrix(self):
        rx_count = 0
        ry_count = 0
        rz_count = 0
        s_count = 0
        t_count = 0
        for idx, operation in enumerate(self.order):
            if operation == 'rx':
                self.M = self.M @ self.rx_mat[rx_count].T
                rx_count += 1
            elif operation == 'ry':
                self.M = self.M @ self.ry_mat[ry_count].T
                ry_count += 1
            elif operation == 'rz':
                self.M = self.M @ self.rz_mat[rz_count].T
                rz_count += 1
            elif operation == 's':
                self.M = self.M @ self.s_mat[s_count].T
                s_count += 1
            elif operation == 't':
                self.M = self.M @ self.t_mat[t_count].T
                t_count += 1
            else:
                raise TransformationError(f'Invalid value \'{operation}\' found in 3-D transformation order '
                                          f'(must be \'rx\', \'ry\', \'rz\', \'s\', or \'t\'')

    def transform(self, coordinates: np.ndarray):
        """Computes the transformation of the coordinates.

        Parameters
        ==========
        coordinates: np.ndarray
          Size N x 3, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return (np.column_stack((coordinates, np.ones(len(coordinates)))) @ self.M)[:, :3]  # x' = x * M


class TransformationError(Exception):
    pass


class AirfoilTransformation:
    """Convenience class for computing transformations of coordinate pairs :math:`(x,y)` to and from the
    airfoil-relative coordinate system."""
    def __init__(self, dx, dy, alf, c):
        self.transform_abs_obj = Transformation2D(tx=[dx], ty=[dy], r=[-alf], sx=[c], sy=[c],
                                                  rotation_units='rad', order='r,s,t')
        self.transform_rel_obj = Transformation2D(tx=[-dx], ty=[-dy], r=[alf], sx=[1.0 / c], sy=[1.0 / c],
                                                  rotation_units='rad', order='t,s,r')

    def transform_abs(self, coordinates: np.ndarray):
        """Computes the transformation of the coordinates from the airfoil-relative coordinate system to the absolute
        coordinate system.

        Parameters
        ==========
        coordinates: np.ndarray
          Size N x 2, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return self.transform_abs_obj.transform(coordinates)

    def transform_rel(self, coordinates: np.ndarray):
        """Computes the transformation of the coordinates from the absolute coordinate system to the airfoil-relative
        coordinate system.

        Parameters
        ==========
        coordinates: np.ndarray
          Size N x 2, where N is the number of coordinates. The columns represent :math:`x` and :math:`y`.
        """
        return self.transform_rel_obj.transform(coordinates)


if __name__ == '__main__':
    coords = np.array([[1, 0], [2, 1], [3, 3], [2, 0]])
    transform2d = Transformation2D(tx=[1], ty=[1], r=[90], sx=[2], sy=[2], rotation_units='deg', order='r,s,t')
    new_coords = transform2d.transform(coords)
    print(f"new_coords = {new_coords}")
    coords_3d = np.array([[0, 0, 0], [1, 2, 3], [-1, 0, -6], [-6, 0, 0]])
    transform3d = Transformation3D(tx=[1], ty=[2], tz=[3], sx=[2], sy=[2], sz=[2], rx=[30], ry=[40], rz=[-180],
                                   rotation_units='deg', order='rx,rz,ry,s,t')
    new_coords_3d = transform3d.transform(coords_3d)



import os

import numpy as np

from pymead import RESOURCE_DIR

BERLIN = np.loadtxt(os.path.join(RESOURCE_DIR, "cmcrameri", "berlin.txt"))
VIK = np.loadtxt(os.path.join(RESOURCE_DIR, "cmcrameri", "vik.txt"))



import numpy as np

from pymead.plugins.IGES.entity import Entity
from pymead.plugins.IGES.iges_param import IGESParam


class RationalBSplineIGES(Entity):
    """
    IGES Entity #126
    """
    def __init__(self, knots: np.ndarray, weights: np.ndarray,
                 control_points_XYZ: np.ndarray, degree: int, start_parameter_value=0.0, end_parameter_value=1.0,
                 unit_normal_x=0.0, unit_normal_y=0.0, unit_normal_z=0.0, planar_flag: bool = False,
                 closed_flag: bool = False, polynomial_flag: bool = False, periodic_flag: bool = False):
        self.upper_index_of_sum = control_points_XYZ.shape[0] - 1
        self.degree = degree
        self.flag1 = int(planar_flag)
        self.flag2 = int(closed_flag)
        self.flag3 = int(polynomial_flag)
        self.flag4 = int(periodic_flag)
        self.knots = knots
        self.weights = weights
        self.control_points = control_points_XYZ
        self.v0 = start_parameter_value
        self.v1 = end_parameter_value
        self.XN = unit_normal_x
        self.YN = unit_normal_y
        self.ZN = unit_normal_z
        parameter_data = [
            IGESParam(self.upper_index_of_sum, "int"),
            IGESParam(self.degree, "int"),
            IGESParam(self.flag1, "int"),
            IGESParam(self.flag2, "int"),
            IGESParam(self.flag3, "int"),
            IGESParam(self.flag4, "int"),
            *[IGESParam(k, "real") for k in self.knots],
            *[IGESParam(w, "real") for w in self.weights],
            *[IGESParam(xyz, "real") for xyz in self.control_points.flatten()],
            IGESParam(self.v0, "real"),
            IGESParam(self.v1, "real"),
            IGESParam(self.XN, "real"),
            IGESParam(self.YN, "real"),
            IGESParam(self.ZN, "real"),
        ]
        super().__init__(126, parameter_data)


class BezierIGES(RationalBSplineIGES):
    def __init__(self, control_points_XYZ: np.ndarray, start_parameter_value=0.0, end_parameter_value=1.0,
                 unit_normal_x=0.0, unit_normal_y=0.0, unit_normal_z=0.0, planar_flag: bool = False,
                 closed_flag: bool = False, periodic_flag: bool = False):
        order = len(control_points_XYZ)
        degree = order - 1
        knots = np.concatenate((np.zeros(order), np.ones(order)))
        weights = np.ones(order)
        polynomial_flag = True
        super().__init__(knots=knots, weights=weights, control_points_XYZ=control_points_XYZ, degree=degree,
                         start_parameter_value=start_parameter_value, end_parameter_value=end_parameter_value,
                         unit_normal_x=unit_normal_x, unit_normal_y=unit_normal_y, unit_normal_z=unit_normal_z,
                         planar_flag=planar_flag, closed_flag=closed_flag, polynomial_flag=polynomial_flag,
                         periodic_flag=periodic_flag)



import signal

import psutil


def kill_child_processes(parent_pid: int, sig: int = signal.SIGTERM):
    """
    Kills all child processes (using ``SIGTERM``) of a process with a given PID.
    Most code is from `this StackOverflow answer <https://stackoverflow.com/a/17112379>`_.

    Parameters
    ----------
    parent_pid: int
        Process ID of the parent

    sig: int
        Signal to send (``SIGTERM`` by default)

    Returns
    -------

    """
    try:
        parent = psutil.Process(parent_pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children(recursive=True)
    for process in children:
        try:
            process.send_signal(sig)
        except psutil.NoSuchProcess:
            continue

    # In the rare case that an instance of XFOIL or MSES does not get shut down, force-close any processes with those
    # names
    for proc in psutil.process_iter():
        if proc.name() in ["xfoil.exe", "mses.exe", "xfoil", "mses"]:
            try:
                proc.send_signal(sig)
            except psutil.NoSuchProcess:
                continue



from PyQt5 import QtWidgets


class QHSeperationLine(QtWidgets.QFrame):
    """From LegendaryCodingNoob @ https://stackoverflow.com/a/61389578"""
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.setMinimumWidth(1)
        self.setFixedHeight(20)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)


class QVSeperationLine(QtWidgets.QFrame):
    """From LegendaryCodingNoob @ https://stackoverflow.com/a/61389578"""
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.setFixedWidth(20)
        self.setMinimumHeight(1)
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)



import os
import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numba import jit
from numpy import zeros, pi, arctan2, sin, cos, sqrt, log, dot
from numpy.linalg import inv

from pymead import DATA_DIR


@jit(nopython=True, cache=True)
def single_element_inviscid(coord: np.ndarray, alpha: float or np.float64):
    r"""
    A linear strength vortex panel method for the inviscid solution of a single airfoil, sped up using the
    just-in-time compiler in *numba*. Directly adapted from "Program 7" of Appendix D in [1].

    [1] J. Katz and A. Plotkin, Low-Speed Aerodynamics, Second Edition, 2nd ed. New York, NY,
    USA: Cambridge University Press, 2004. Accessed: Mar. 07, 2023. [Online].
    Available: `<https://asmedigitalcollection.asme.org/fluidsengineering/article/126/2/293/458666/LowSpeed-Aerodynamics-Second-Edition>`_

    Parameters
    ==========
    coord: np.ndarray
      An :math:`N \times 2` array of airfoil coordinates, where :math:`N` is the number of coordinates, and the columns
      represent :math:`x` and :math:`y`

    alpha: float or np.float64
      Angle of attack of the airfoil

    Returns
    =======
    np.ndarray, np.ndarray, float
      The first returned array is of size :math:`(N-1) \times 2` and represents the :math:`x` and :math:`y` locations
      of the collocation points, where :math:`N` is the number of airfoil coordinates. The second returned array is a
      one-dimensional array with length :math:`(N-1)` representing the surface pressure coefficient at each collocation
      point. The final returned value is the lift coefficient.
    """

    N = len(coord)          # Number of panel end points
    M = N - 1                 # Number of control points

    EP = zeros((N, 2))      # Clockwise-defined panel end points
    EPT = zeros((N, 2))     # End points from file
    PT1 = zeros((M, 2))     # Start point of panel
    PT2 = zeros((M, 2))     # End point of panel
    CO = zeros((M, 2))      # Collocation point
    A = zeros((N, N))       # Aerodynamic influence coefficient matrix
    B = zeros((N, N))       # Tangential induced velocities (with gammas)
    TH = zeros((M,))      # Panel angle
    DL = zeros((M,))      # Panel length
    RHS = zeros((N, 1))     # Freestream component normal to panel
    V = zeros((M,))       # Panel tangential velocity

    ALPHA = alpha  # AOA in deg
    AL = ALPHA * pi / 180

    EPT[:, 0] = coord[:, 0]    # Read in x/c position of panel end points
    EPT[:, 1] = coord[:, 1]    # Read in y/c position of panel end points

    # Order panel end points defined in clockwise direction
    for i in range(N):
        EP[i, 0] = EPT[N - i - 1, 0]
        EP[i, 1] = EPT[N - i - 1, 1]

    # Define end points of each panel (PT1 is beginning point, PT2 is end point)
    for i in range(M):
        PT1[i, 0] = EP[i, 0]
        PT2[i, 0] = EP[i+1, 0]
        PT1[i, 1] = EP[i, 1]
        PT2[i, 1] = EP[i+1, 1]

    # Determine local slope of each panel
    for i in range(M):
        DZ = PT2[i, 1] - PT1[i, 1]
        DX = PT2[i, 0] - PT1[i, 0]
        TH[i] = arctan2(DZ, DX)

    # Identify collocation points for each panel (half-panel location)
    for i in range(M):
        CO[i, 0] = (PT2[i, 0] - PT1[i, 0]) / 2 + PT1[i, 0]
        CO[i, 1] = (PT2[i, 1] - PT1[i, 1]) / 2 + PT1[i, 1]

    # Determine influence coefficients
    for i in range(M):
        for j in range(M):

            # Determine location of collocation point i in terms of panel j
            # coordinates
            XT = CO[i, 0] - PT1[j, 0]
            ZT = CO[i, 1] - PT1[j, 1]
            X2T = PT2[j, 0] - PT1[j, 0]
            Z2T = PT2[j, 1] - PT1[j, 1]

            X = XT*cos(TH[j]) + ZT*sin(TH[j])
            Z = -XT*sin(TH[j]) + ZT*cos(TH[j])
            X2 = X2T*cos(TH[j]) + Z2T*sin(TH[j])
            Z2 = 0

            # Store length of each panel (only required for first loop in i)
            if i == 0:
                DL[j] = X2

            # Determine radial distance and angle between corner points of jth
            # panel and ith control point
            R1 = sqrt(X**2 + Z**2)
            R2 = sqrt((X - X2)**2 + Z**2)
            TH1 = arctan2(Z, X)
            TH2 = arctan2(Z, X - X2)

            # Determine influence coefficient of jth panel on ith control point
            # (include consideration for self-induced velocities)
            if i == j:
                U1L = -0.5*(X - X2) / X2
                U2L = 0.5*X / X2
                W1L = -0.15916
                W2L = 0.15916
            else:
                U1L = -(Z*log(R2/R1) + X*(TH2 - TH1) - X2*(TH2 - TH1)) / (6.28319*X2)
                U2L = (Z*log(R2/R1) + X*(TH2 - TH1)) / (6.28319*X2)
                W1L = -((X2 - Z*(TH2 - TH1)) - X*log(R1/R2) + X2*log(R1/R2)) / (6.28319*X2)
                W2L = ((X2 - Z*(TH2 - TH1)) - X*log(R1/R2)) / (6.28319*X2)

            # Rotate coordinates back from jth panel reference frame to airfoil
            # chord frame
            U1 = U1L * np.cos(-TH[j]) + W1L * np.sin(-TH[j])
            U2 = U2L*cos(-TH[j]) + W2L*sin(-TH[j])
            W1 = -U1L*sin(-TH[j]) + W1L*cos(-TH[j])
            W2 = -U2L*sin(-TH[j]) + W2L*cos(-TH[j])

            # Define AIC: A(i,j) is the component of velocity normal to control
            # point i due to panel j
            # B(i,j) is the tangential velocity along control point i due to
            # panel j, used after solving for gammas
            if j == 0:
                A[i, 0] = -U1*sin(TH[i]) + W1*cos(TH[i])
                HOLDA = -U2*sin(TH[i]) + W2*cos(TH[i])
                B[i, 0] = U1*cos(TH[i]) + W1*sin(TH[i])
                HOLDB = U2*cos(TH[i]) + W2*sin(TH[i])
            elif j == M - 1:
                A[i, M - 1] = -U1*sin(TH[i]) + W1*cos(TH[i]) + HOLDA
                A[i, N - 1] = -U2*sin(TH[i]) + W2*cos(TH[i])
                B[i, M - 1] = U1*cos(TH[i]) + W1*sin(TH[i]) + HOLDB
                B[i, N - 1] = U2*cos(TH[i]) + W2*sin(TH[i])
            else:
                A[i, j] = -U1*sin(TH[i]) + W1*cos(TH[i]) + HOLDA
                HOLDA = -U2*sin(TH[i]) + W2*cos(TH[i])
                B[i, j] = U1*cos(TH[i]) + W1*sin(TH[i]) + HOLDB
                HOLDB = U2*cos(TH[i]) + W2*sin(TH[i])

        # Set up freestream component of boundary condition
        RHS[i, 0] = cos(AL)*sin(TH[i]) - sin(AL)*cos(TH[i])

    # Enforce Kutta condition
    RHS[N - 1, 0] = 0

    A[N - 1, 0] = 1
    A[N - 1, N - 1] = 1

    # Invert A matrix to solve for gammas
    G = dot(inv(A), RHS)

    # With known gammas, solve for CL, CPs
    CL = 0.0

    for i in range(M):
        VEL = 0.0
        for j in range(N):
            VEL = VEL + (B[i, j] * G[j])[0]
        V[i] = VEL + cos(AL)*cos(TH[i]) + sin(AL)*sin(TH[i])
        CL = CL + ((G[i] + G[i + 1]) * DL[i])[0]

    CP = 1 - V**2

    return CO, CP, CL


def _main():
    # Insert airfoil points here
    coords = pd.read_csv(os.path.join(DATA_DIR, 'naca0012.dat'), skiprows=1, names=['x', 'y'], delim_whitespace=True)
    # print(coord)
    coords = coords.to_numpy()
    single_element_inviscid(coords, 4.0)
    t1 = time.time()
    CO, CP, CL = single_element_inviscid(coords, 4.0)
    t2 = time.time()
    print(f"calculation time = {t2 - t1:.4e} seconds")

    fig, axs = plt.subplots()
    axs.plot(CO[:, 0], CP)
    axs.set_xlabel(r"$x/c$")
    axs.set_ylabel(r"$C_p$")
    axs.invert_yaxis()
    print(f"CL = {CL:.4g}")
    plt.show()


if __name__ == '__main__':
    _main()




import os

import pyqtgraph as pg
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QToolBar, QToolButton

from pymead import ICON_DIR, GUI_SETTINGS_DIR, q_settings
from pymead.utils.read_write_files import load_data


class MainIconToolbar(QToolBar):
    """Class containing the ToolBar buttons for the GUI. Note that the button settings are loaded from a JSON files
    stored in the gui directory and stored as a dict. The actual QToolButtons themselves are stored inside the
    \'buttons\' attribute."""
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        self.new_airfoil_location = None
        self.symmetry_dialog = None
        self.pos_constraint_dialog = None
        self.icon_dir = ICON_DIR
        self.parent.addToolBar(self)
        self.button_settings = load_data(os.path.join(GUI_SETTINGS_DIR, "buttons.json"))
        self.button_settings.pop("template")
        self.buttons = None
        self.browser = None
        self.add_all_buttons()

    def add_all_buttons(self):
        self.buttons = {}
        for button_name, button_settings in self.button_settings.items():
            self.buttons[button_name] = {}
            # Add the icon:
            self.buttons[button_name]["icon"] = QIcon(os.path.join(self.icon_dir, button_settings["icon"]))

            # Set up the physical button:
            button = QToolButton(self)
            self.buttons[button_name]["button"] = button
            button.setStatusTip(button_settings["status_tip"])  # add the status tip
            button.setCheckable(button_settings["checkable"])  # determine whether the button can be checked
            button.setIcon(self.buttons[button_name]["icon"])  # set the icon for the button
            if button_settings["checkable"]:
                button.toggled.connect(getattr(self, button_settings["function"]))  # add the functionality
            else:
                button.clicked.connect(getattr(self, button_settings["function"]))  # add the functionality
            if button.isCheckable():
                if button_settings["checked-by-default"]:
                    button.toggle()
            self.addWidget(button)

    def on_stop_button_pressed(self):
        self.parent.stop_process()

    def on_grid_button_pressed(self):
        self.parent.airfoil_canvas.toggleGrid()
        for dw in self.parent.dock_widgets:
            if isinstance(dw.widget(), pg.GraphicsLayoutWidget):
                v = dw.widget().getItem(0, 0)
                x_state = v.ctrl.xGridCheck.checkState()
                y_state = v.ctrl.yGridCheck.checkState()
                if x_state or y_state:
                    v.showGrid(x=False, y=False)
                else:
                    v.showGrid(x=True, y=True)
            elif hasattr(dw.widget(), "toggle_grid"):
                dw.widget().toggle_grid()

    def change_background_color_button_toggled(self, checked):

        if checked:
            self.parent.set_theme("dark")
            q_settings.setValue("dark_theme_checked", 2)
        else:
            self.parent.set_theme("light")
            q_settings.setValue("dark_theme_checked", 0)

        self.parent.show()

    def on_draw_points_pressed(self):
        self.parent.airfoil_canvas.drawPoints()

    def on_draw_lines_pressed(self):
        self.parent.airfoil_canvas.drawLines()

    def on_draw_bezier_pressed(self):
        self.parent.airfoil_canvas.drawBeziers()

    def on_generate_airfoil_pressed(self):
        self.parent.airfoil_canvas.generateAirfoil()

    def on_generate_web_airfoil_pressed(self):
        self.parent.airfoil_canvas.generateWebAirfoil()

    def on_generate_mea_pressed(self):
        self.parent.airfoil_canvas.generateMEA()

    def on_add_distance_constraint_pressed(self):
        self.parent.airfoil_canvas.addDistanceConstraint()

    def on_add_rel_angle_constraint_pressed(self):
        self.parent.airfoil_canvas.addRelAngle3Constraint()

    def on_add_perp_constraint_pressed(self):
        self.parent.airfoil_canvas.addPerp3Constraint()

    def on_add_anti_parallel_constraint_pressed(self):
        self.parent.airfoil_canvas.addAntiParallel3Constraint()

    def on_add_symmetry_constraint_pressed(self):
        self.parent.airfoil_canvas.addSymmetryConstraint()

    def on_add_roc_constraint_pressed(self):
        self.parent.airfoil_canvas.addROCurvatureConstraint()

    def on_help_button_pressed(self):
        self.parent.show_help()
